import numpy as np
from itertools import product

import loter.initparam as initparam
import loter.initdata as initdata
import loter.opti as opti
import loter.estimatea as esta
import loter.estimateh as esth
import loter.graph as ests
import loter.predcombine as combiner
import loter.metrics as metrics
"""
A Pipeline is a function taking the matrix Genotype in input and
return some results.
"""
def build_pipeline(init_param, init_data, opti):
    def pipeline(G, **kwargs):
        param = init_param(G, **kwargs)
        data = initdata.create_data(G, param)
        data = init_data(data, param)
        res = opti(data, param)
        return res

    return pipeline

class Pipeline(object):

    def __init__(self, init_param, init_data, opti):
        self.init_param = init_param
        self.init_data = init_data
        self.opti = opti

    def __call__(self, G, **kwargs):
        self.pipeline = build_pipeline(self.init_param,
                                       self.init_data,
                                       self.opti)

        return self.pipeline(G, **kwargs)

    def run(self, G, nbrun=10, **kwargs):
        l_res = []

        self.pipeline = build_pipeline(self.init_param,
                                       self.init_data,
                                       self.opti)

        for i in range(nbrun):
            l_res.append(self.pipeline(G, **kwargs))

        return l_res

class PipelineCrossValidate(Pipeline):

    def __call__(self, G, **kwargs):
        param_initializer = self.init_param(G, **kwargs)
        self.pipeline = build_pipeline(param_initializer,
                                       self.init_data,
                                       self.opti)

        return self.pipeline(G, **kwargs)

    def run(self, G, nbrun=10, **kwargs):
        l_res = []

        param_initializer = self.init_param(G, **kwargs)
        self.pipeline = build_pipeline(param_initializer,
                                       self.init_data,
                                       self.opti)

        for i in range(nbrun):
            print("Run {0}".format(i))
            l_res.append(self.pipeline(G, **kwargs))

        return l_res

def cross_validate_param(G, p_sampling=0.1, w_h=100, nb_iter=20, **kwargs):

    (n, m) = G.shape
    param_cr = {
        "w_h": w_h,
        "nb_iter": nb_iter
    }

    for key, value in list(kwargs.items()):
        param_cr[key] = value


    grid = list(product(
        np.arange(1.0, 3.0, 0.2),
        np.arange(5, max(10, n/2), 10),
        (param_cr["w_h"],),
        (param_cr["nb_iter"],)
    ))

    l_error = []
    for penalty, nbclust, w_h, nb_iter in grid:
        param = {"penalty": penalty,
                "nbclust": nbclust,
                "w_h": w_h,
                "nb_iter": nb_iter}

        G_in = np.copy(G[:,:min(n,5000)])
        G_nan = np.copy(G_in)
        selected_idx = np.where(G_nan != 3)
        sample_idx = np.random.choice([True, False],
                                      size=len(selected_idx[0]),
                                      p=[p_sampling, 1.0-p_sampling])
        masked_idx = (selected_idx[0][sample_idx], selected_idx[1][sample_idx])
        G_nan[masked_idx] = 3
        G_nan.astype(np.uint8)

        method = pipelines["classic_pipeline"]
        l_res = method.run(G_nan, nbrun=20, **param)
        G_res = combiner.combiner_G["G vote"](l_res)
        error = metrics.imputation_error_rate(G_res, G_in, G_nan)
        l_error.append((param, error))
        #print(error, param)

    param_min = min(l_error, key = lambda t: t[1])[0]

    #print("Selected Parameters : ")
    #print(param_min)

    def init_param(G, **kwargs):
        param = param_min

        for key, value in list(kwargs.items()):
            param[key] = value

        return param

    return init_param

def applied_cross_validate_param(G, **kwargs):
    return cross_validate_param(G, **kwargs)(G, **kwargs)

classic_pipeline = Pipeline(
    initparam.param_initializers["classic_init"],
    initdata.data_initializers["rand"],
    opti.opti["classic_opti"]
)

tree_pipeline = Pipeline(
    initparam.param_initializers["classic_init"],
    initdata.data_initializers["a tree"],
    opti.opti["classic_opti"]
)

tree_join_pipeline = Pipeline(
    initparam.param_initializers["classic_init"],
    initdata.data_initializers["a tree"],
    opti.opti["join_opti"]
)

tree_oldh_pipeline = Pipeline(
    initparam.param_initializers["classic_init"],
    initdata.data_initializers["a tree"],
    opti.opti["opti_h_old"]
)

classic_pipeline_prob = Pipeline(
    initparam.param_initializers["classic_init"],
    initdata.data_initializers["rand"],
    opti.opti_build(
        esth.optih["opti_prob"],
        esta.optimize_A,
        ests.optimize_S
    )
)

pipeline_old_h = Pipeline(
    initparam.param_initializers["classic_init"],
    initdata.data_initializers["rand"],
    opti.opti["opti_h_old"]
)

classic_crossvalidate_pipeline = PipelineCrossValidate(
    cross_validate_param,
    initdata.data_initializers["rand"],
    opti.opti["classic_opti"]
)

classic_crossvalidate_ag_pipeline = PipelineCrossValidate(
    cross_validate_param,
    initdata.data_initializers["a with g, h rand"],
    opti.opti["classic_opti"]
)

join_ag_pipeline = Pipeline(
    initparam.param_initializers["classic_init"],
    initdata.data_initializers["a with g, h rand"],
    opti.opti["join_opti"]
)

join_then_classic_pipeline = PipelineCrossValidate(
    cross_validate_param,
    initdata.data_initializers["a with g, h rand"],
    opti.opti["join_then_classic"]
)

join_then_classic_pipeline_withsampling = PipelineCrossValidate(
    cross_validate_param,
    initdata.data_initializers["a with g, h rand"],
    opti.opti["join_then_classic_withsampling"]
)

join_then_classic_pipeline_initkmeans = PipelineCrossValidate(
    cross_validate_param,
    initdata.data_initializers["kmeans"],
    opti.opti["join_then_classic"]
)

join_initkmeans_balance = Pipeline(
    initparam.param_initializers["classic_init"],
    initdata.data_initializers["kmeans"],
    opti.opti["join_balance"]
)

classic_crosseach_pipeline = Pipeline(
    applied_cross_validate_param,
    initdata.data_initializers["rand"],
    opti.opti["classic_opti"]
)

pipelines = {
    "classic_pipeline": classic_pipeline,
    "pipeline_old_h": pipeline_old_h,
    "crossvalidate": classic_crossvalidate_pipeline,
    "cross_ag": classic_crossvalidate_ag_pipeline,
    "join_ag": join_ag_pipeline,
    "join_then_classic": join_then_classic_pipeline,
    "join_then_classic_withsampling": join_then_classic_pipeline_withsampling,
    "join_then_classic_initkmeans": join_then_classic_pipeline_initkmeans,
    "join_initkmeans_balance": join_initkmeans_balance,
    "crosseachvalidate": classic_crosseach_pipeline
}



import numpy as np

class Node(object):

    def __init__(self, depth, value):
        self.depth = depth
        self.value = value
        self.left = None
        self.right = None

    def __repr__(self):
        return str(self.value)

def all_paths(node):
    if node.left is None and node.right is None:
        return [[node.value]]
    elif node.left is None:
        res = all_paths(node.right)
        for e in res:
            e.insert(0, node.value)
        return res
    elif node.right is None:
        res = all_paths(node.left)
        for e in res:
            e.insert(0, node.value)
        return res
    else:
        res_left = all_paths(node.left)
        res_right = all_paths(node.right)
        for e in res_left:
            e.insert(0, node.value)
        for e in res_right:
            e.insert(0, node.value)
        return res_left + res_right

def npy_tree(tree):
    res = np.array(all_paths(tree))
    return res[:, 1:].astype(np.uint8)

def update_left(l_node, i):
    node_added = list()
    for node in l_node:
        if node is None:
            print("Warning")
        else:
            if node.left is None:
                node.left = Node(i, 0)

            node_added.append(node.left)

    return node_added

def update_right(l_node, i):
    node_added = list()
    for node in l_node:
        if node is None:
            print("Warning")
        else:
            if node.right is None:
                node.right = Node(i, 1)

            node_added.append(node.right)

    return node_added

def update_tree(root, arr):
    curr = root
    to_update = [curr]
    for i, elem in enumerate(arr):
        nextlevel = list()
        if elem == 0:
            new_nodes = update_left(to_update, i)
            nextlevel.extend(new_nodes)
        elif elem == 2:
            new_nodes = update_right(to_update, i)
            nextlevel.extend(new_nodes)
        else:
            new_nodes = update_left(to_update, i)
            nextlevel.extend(new_nodes)
            new_nodes = update_right(to_update, i)
            nextlevel.extend(new_nodes)

        to_update = nextlevel

def build_tree(mat):
    root = Node(0, "root")
    for row in mat:
        update_tree(root, row)

    return root

def count_width(tree):
    if tree is None:
        return 0
    else:
        if tree.left is None and tree.right is None:
            return 1
        else:
            return count_width(tree.left) + count_width(tree.right)

def arr_to_h(k, arr):
    res_h = []
    curr_pos = 1
    last_pos = 0
    n, m = arr.shape
    while curr_pos <= m:
        width = 0
        while width < k and curr_pos <= m:
            t = build_tree(arr[:, last_pos:curr_pos])
            width = count_width(t)
            curr_pos = curr_pos + 1
        h = npy_tree(t)
        if len(h) < k:
            nb_rows = k - len(h)
            h = np.vstack([h, h[np.random.randint(h.shape[0], size=nb_rows), :]])
        h = np.random.shuffle(h)
        res_h.append(h)
        last_pos = curr_pos - 1

    print(res_h)
    return np.hstack(res_h)

def mat_split(k, arr, pos):
    actual_pos = pos
    arr_fp = np.fliplr(arr[:, :pos])
    arr_lp = arr[:, pos:]

    res_fp = np.fliplr(arr_to_h(k, arr_fp))
    res_lp = arr_to_h(k, arr_lp)
    print(res_fp.shape)
    print(res_lp.shape)

    return np.hstack([res_fp, res_lp])




import numpy as np
import functools
import operator
import loter.utils as utils

def unpack_data_from_input(l_input):
    return [data for data, param in l_input]

def compress_H(H):
    H1, H2 = np.copy(H[::2]).astype(np.int), np.copy(H[1::2]).astype(np.int)
    H1[H1 == 1] = 4
    H2[H2 == 1] = 5
    return H1 + H2

def build_H1(H):
    r = np.copy(H)
    r[r == 4] = 1
    r[r == 9] = 1
    r[r == 5] = 0
    return r

def build_H2(H):
    r = np.copy(H)
    r[r == 4] = 0
    r[r == 9] = 1
    r[r == 5] = 1
    return r

def pipeline_H_mean(l_H, heter):
    def treat_H(H):
        H_res = compress_H(H)
        homozygous = ~heter
        H_res[homozygous] = 0
        # compute if the switches by compute differences
        H_res[heter] = np.absolute(np.ediff1d(H_res[heter], to_begin=0))
        # set the first heterozygous to 0
        H_res[range(H_res.shape[0]), np.argmax(heter, axis=1)] = 0
        # For the accumulator with set the non-switches to -1 instead of 0
        H_res[np.logical_and(heter, H_res == 0)] = -1
        return H_res
    l_H_treated = map(treat_H, l_H)
    hn, hm = l_H[0].shape
    acc = functools.reduce(operator.add, l_H_treated, np.zeros((hn/2, hm), dtype=int))

    # Prepare the result for the output
    acc[np.logical_and(acc >= 0, heter)] = 1
    acc[np.logical_and(acc < 0, heter)] = 0
    acc[heter] = np.cumsum(acc[heter]) % 2
    acc[np.logical_and(acc == 1, heter)] = 5
    acc[np.logical_and(acc == 0, heter)] = 4

    result = np.copy(l_H[0])
    result[::2][heter] = build_H1(acc)[heter]
    result[1::2][heter] = build_H2(acc)[heter]

    return result

def pipeline_H_weighted_mean(l_SA, heter):
    l_mat_p10 = []
    for SA in l_SA:
        A1, A2 = SA[::2] + SA[1::2]
        mat_p10 = A1 * (1 - A2) / (A1 * (1 - A2) + A2 * (1 - A1))

def combine_H_mean(l_input):
    l_data = unpack_data_from_input(l_input)
    l_H = [elem["H"] for elem in l_data]
    G = l_data[0]["G"]
    heter = (G == 1)
    return pipeline_H_mean(l_H, heter)

def combine_G_mean(l_input):
    l_data = unpack_data_from_input(l_input)
    l_H = [elem["H"] for elem in l_data]
    list_G = [H[::2] + H[1::2] for H in l_H]
    mean_G = np.mean(list_G, axis=(0))
    return np.around(mean_G)

def combine_G_mode(l_input):
    l_data = unpack_data_from_input(l_input)
    l_H = [elem["H"] for elem in l_data]
    list_G = [H[::2] + H[1::2] for H in l_H]
    mode_G = utils.mode(np.dstack(list_G), axis=2)[0]
    res = np.reshape(mode_G, mode_G.shape[:-1])
    return res.astype(np.uint8)

combiner_H = {
    "H mean": combine_H_mean
}
combiner_G = {
    "G mean": combine_G_mean,
    "G vote": combine_G_mode
}



import ctypes as C
import os
import sys
import pkg_resources
from glob import glob

class LoterLibraryNotFound(Exception):
    """Error thrown when loter is not found"""
    pass

def _get_lib():
    # dirpath = os.path.dirname(os.path.abspath(os.path.expanduser(__file__)))
    # libpath_possible = [os.path.join(dirpath, '../../lib/', 'libloter.so'),
    #                     os.path.join(dirpath, '../', 'libloter.so')]
    # libpath = [path for path in libpath_possible if os.path.exists(path) and os.path.isfile(path)]
    # if not libpath:
    #     raise LoterLibraryNotFound("Library not found in :\n" + '\n'.join(libpath_possible))

    libpath = glob(pkg_resources.resource_filename('loter',
                   os.path.join('_loter_wrap*.so')))

    if not libpath:
        for dir in sys.path:
            libpath = glob(os.path.join(dir, 'loter', '_loter_wrap*.so'))
            if libpath:
                break
        if not libpath:
            raise LoterLibraryNotFound("Library `_loter_wrap.so` not found")

    selected_lib = libpath[0]
    lib = C.cdll.LoadLibrary(selected_lib)

    return lib

_LIB = _get_lib()



import numpy as np
import scipy.stats as stats
import pandas as pd

import loter.pipeline as lt
import loter.initparam as initparam
import loter.initdata as initdata
import loter.opti as opti
import loter.estimatea as esta
import loter.estimateh as esth
import loter.graph as ests

##################################################################
#                                                                #
# Pipeline to optimize H (admixed haplotypes) and S (selection)  #
# given A (ancestral haplotypes) simultaneously                  #
#                                                                #
# Initialization:                                                #
# - init A with ancestral haplotypes                             #
#                                                                #
# Optimization:                                                  #
# - join optimisation of H and S                                 #
#                                                                #
# Computational Complexity:                                      #
# - n, the number of ancestral individuals                       #
# - m, the number of SNPs                                        #
# complexity -> O(n^2 * m)                                       #
#                                                                #
# Remarks:                                                       #
# In practice the pipeline is use as a phase corrector module    #
##################################################################

def init_fix_a(data, param):
    def init_a_fix(A, G):
        return param["A_in"]
    return initdata.init_data(data, init_a_fix, initdata.init_h_rand)

def opti_A_fix_join(data, param):
    data, param = ests.optimize_SHknn(data, param)
    data, param = esth.optimize_H_old(data, param)

    return data, param

fixa_pip_A_join = lt.Pipeline(
    initparam.param_initializers["classic_init"],
    init_fix_a,
    opti_A_fix_join,
)

##################################################################
#                                                                #
# Pipeline to optimize S (selection) given A                     #
# (ancestral haplotypes) and H (admixed haplotypes)              #
#                                                                #
# Initialization:                                                #
# - init A with ancestral haplotypes                             #
# - init H with admixed haplotypes                               #
#                                                                #
# Optimization:                                                  #
# - optimisation of S                                            #
#                                                                #
# Computational Complexity:                                      #
# - n, the number of ancestral individuals                       #
# - m, the number of SNPs                                        #
# complexity -> O(n * m)                                         #
#                                                                #
##################################################################

def init_fix_ah(data, param):
    def init_a_fix(A, G):
        return param["A_in"]
    def init_h_fix(H, G):
        return param["H_in"]
    return initdata.init_data(data, init_a_fix, init_h_fix)

def opti_AH_fix_knn(data, param):

    data["A"] = data["A"].astype(np.uint8)
    data["S"] = data["S"].astype(np.uint32)
    data, param = ests.optimize_Sknn(data, param)

    return data, param

fixa_pip_AH_knn = lt.Pipeline(
    initparam.param_initializers["classic_init"],
    init_fix_ah,
    opti_AH_fix_knn
)

def learn_Sknn(pop, A_in, H_in, weights, penalty=40, num_threads=10):
    G_pop = pop["G"]
    H_pop = H_in

    l_res_mix = fixa_pip_AH_knn(G_pop,
                                nb_iter=1, nbclust=len(A_in), penalty=penalty,
                                num_threads=num_threads,
                                weights=weights,
                                A_in=A_in,
                                H_in=H_pop
    )

    return l_res_mix[0]["S"]

def learn_S_join(pop, A_in, penalty=40, small_penalty=0, num_threads=10):
    G_pop = pop["G"]

    l_res_mix = fixa_pip_A_join(G_pop,
                                nb_iter=1, nbclust=len(A_in), penalty=penalty,
                                A_in=A_in,
                                small_penalty=small_penalty,
                                num_threads=num_threads
    )

    return l_res_mix[0]["S"], l_res_mix[0]["H"]

def get_items(dict_object):
    """
    Compatible Python 2 et 3 get item for dictionnary
    """
    for key in dict_object:
        yield key, dict_object[key]

def clusters_to_list_pop(S, l_k):
    """
    From a selection matrix S, compute the origin of each SNP.

    input:
    S -- matrix where we are copying
    l_k -- populations sizes
    """
    res = np.copy(S)

    a = np.repeat(np.arange(len(l_k)), l_k)
    b = np.arange(sum(l_k))
    d = {k: v for v, k in zip(a, b)}
    for k, v in get_items(d): res[S==k] = v
    return res

def locanc_g_knn(l_h, g_adm, penalty=40, small_penalty=0, num_threads=10):
    A_in = np.ascontiguousarray(np.vstack(l_h))
    S_adm, H = learn_S_join({"G": g_adm}, A_in, penalty, small_penalty, num_threads)
    result = clusters_to_list_pop(S_adm, [len(A) for A in l_h])

    return result, S_adm, H

def locanc_h_knn(l_h, h_adm, penalty=40, num_threads=10):
    A_in = np.ascontiguousarray(np.vstack(l_h))
    g_adm = h_adm[::2] + h_adm[1::2]
    n, m = h_adm.shape
    weights = np.ones(m)
    S_adm = learn_Sknn({"G": g_adm, "H": h_adm}, A_in, h_adm, weights, penalty, num_threads)
    result = clusters_to_list_pop(S_adm, [len(A) for A in l_h])

    return result, S_adm

def update_counts(counts, arr, k=2):
    for p in range(k):
        counts[p,:,:][arr == p] += 1
    return counts

def mode(counts):
    argmax = np.argmax(counts, axis=0)
    return argmax, argmax.choose(counts)

def encode_haplo(H):
    H1, H2 = H[::2], H[1::2]
    return ((np.maximum(H1, H2) * (np.maximum(H1, H2) + 1)) / 2) + np.minimum(H1, H2)

def loter_multiple_pops(l_H, h_adm, lambd, num_threads=10, default=True):
    odd = False
    if h_adm.shape[0] % 2 != 0 & default:
        odd = True
        h_adm = np.vstack([h_adm, np.repeat(0, h_adm.shape[1])])

    res_loter, _= locanc_h_knn([h.astype(np.uint8) for h in l_H],
                               h_adm.astype(np.uint8), lambd, num_threads)

    if odd & default:
        res_loter = res_loter[:res_loter.shape[0]-1]

    return res_loter

def boostrap_loter_multiple_pops(l_H, h_adm, lambd, counts, nbrun=20, num_threads=10):

    def shuffle(H):
        n, m = H.shape
        return H[np.random.randint(n, size=n), :]

    if nbrun > 1:
        for i in range(nbrun):
            shuffled_H = [shuffle(h) for h in l_H]
            counts = update_counts(counts,
                                   loter_multiple_pops(shuffled_H,
                                                       h_adm,
                                                       lambd,
                                                       num_threads,
                                                       False),
                                   len(l_H)
            )
    else:
        counts = update_counts(counts,
                               loter_multiple_pops(l_H,
                                                   h_adm,
                                                   lambd,
                                                   num_threads,
                                                   False),
                               len(l_H)
        )

    return counts

def loter_local_ancestry(l_H, h_adm, range_lambda=np.arange(1.5, 5.5, 0.5),
                         rate_vote=0.5, nb_bagging=20, num_threads=10,
                         default=True):

    odd = False
    if h_adm.shape[0] % 2 != 0 & default:
        odd = True
        h_adm = np.vstack([h_adm, np.repeat(0, h_adm.shape[1])])

    input_loter = (l_H, h_adm)
    n, m = h_adm.shape
    counts = np.zeros((len(l_H), n, m))
    for l in range_lambda:
        res_boostrap = boostrap_loter_multiple_pops(*input_loter, lambd=l,
                                                    counts=counts, nbrun=nb_bagging,
                                                    num_threads=num_threads)
    res_tmp = mode(counts)

    if default:
        if odd:
            res_loter = (res_tmp[0][:res_tmp[0].shape[0]-1],
                         res_tmp[1][:res_tmp[1].shape[1]-1])
        else:
            res_loter = res_tmp
        return res_loter
    else:
        r = vote_and_impute(res_tmp, rate_vote)
        return r, res_tmp

def diploid_sim(cluster_found, cluster_truth):
    (n,m) = cluster_found.shape
    return np.count_nonzero(cluster_found == cluster_truth) / float(n*m)

def find_lambda(s_in, threshold = 0.90, min_lambda = 1,
                max_lambda = 500, num_threads=10):
    n, m = s_in.shape
    if max_lambda - min_lambda <= 1:
        return locanc_g_knn([np.zeros((1,m)), np.ones((1,m))],
                            s_in, min_lambda, min_lambda, num_threads)
    else:
        mean = (max_lambda - min_lambda) / 2 + min_lambda
        r_g, s_g, h_g = locanc_g_knn([np.zeros((1,m)), np.ones((1,m))],
                                     s_in, mean, mean, num_threads)
        sim = diploid_sim(r_g[::2] + r_g[1::2], s_in)
        if sim > threshold:
            return find_lambda(s_in, threshold, min_lambda = (max_lambda - min_lambda) / 2 + min_lambda,
                               max_lambda = max_lambda, num_threads=num_threads)
        else:
            return find_lambda(s_in, threshold, min_lambda = min_lambda,
                               max_lambda = max_lambda  - ((max_lambda - min_lambda) / 2),
                               num_threads=num_threads)

def vote_and_impute(s, percent_threshold=0.5):
    def select_val(s, percent_threshold):
        max_s, min_s = np.max(s[1]), np.min(s[1])
        threshold = percent_threshold*(max_s - min_s) + min_s
        select = np.logical_and(s[1][::2] >= threshold,
                                s[1][1::2] >= threshold)
        arr = encode_haplo(s[0])
        arr[np.logical_not(select)] = 255
        return arr

    arr = select_val(s, percent_threshold)

    n, m = arr.shape
    res = np.copy(arr)

    for i in range(n):
        serie = pd.Series(arr[i])
        serie.loc[serie == 255] = np.nan
        try:
            res[i] = serie.dropna().reindex(range(m), method='nearest').values
        except:
            res[i] = arr[i]

    return res

def loter_smooth(l_H, h_adm, range_lambda=np.arange(1.5, 5.5, 0.5),
                 threshold=0.90, rate_vote=0.5, nb_bagging=20, num_threads=10):

    if h_adm.shape[0] % 2 != 0:
        raise ValueError("`loter_smooth` is designed to analyze haplotypes from diploid species, `l_H.shape[0]` should be an even number. In other cases, you can use the function `loter_local_ancestry`.")

    res_impute, res_raw = loter_local_ancestry(l_H, h_adm, range_lambda,
                                               rate_vote, nb_bagging, num_threads,
                                               False)
    result = np.copy(res_impute)
    result_hap = []
    for i in range(len(res_impute)):
        arr_input = np.ascontiguousarray(np.array([res_impute[i]])).astype(np.uint8)
        r, _, _ = find_lambda(arr_input, threshold=threshold, num_threads=num_threads)
        result_hap.append(r)
        result[i] = r[::2] + r[1::2]

    return np.vstack(result_hap)



import numpy as np

import loter.predcombine as pred
import loter.utils as utils

def cost_error_l2(G, H, S, A, lambd, mask=None):
    SA = utils.build_SA(S, A)
    H_minus_SA = H - SA

    fit_error = np.sum(H_minus_SA**2)
    penalty_error = lambd * np.count_nonzero(np.diff(S) != 0)
    return (fit_error, penalty_error)

def switch_error_rate(H_found, H_truth):
    H_found_comp = pred.compress_H(H_found)
    H_truth_comp = pred.compress_H(H_truth)
    heter = np.logical_or(H_truth_comp == 4, H_truth_comp == 5)

    H_found_comp[heter] = np.absolute(np.ediff1d(H_found_comp[heter], to_begin=0))
    H_found_comp[range(H_found_comp.shape[0]), np.argmax(heter, axis=1)] = 0

    H_truth_comp[heter] = np.absolute(np.ediff1d(H_truth_comp[heter], to_begin=0))
    H_truth_comp[range(H_truth_comp.shape[0]), np.argmax(heter, axis=1)] = 0

    not_first_heter = np.copy(heter)
    not_first_heter[range(not_first_heter.shape[0]), np.argmax(heter, axis=1)] = False
    nb_possible_switches = np.count_nonzero(not_first_heter)
    nb_diff = np.count_nonzero(H_found_comp[heter] != H_truth_comp[heter])

    try:
        return nb_diff / float(nb_possible_switches)
    except ZeroDivisionError:
        return 0.0

def imputation_error_rate(G_found, G_truth, G_nan, mask=None):
    if (not(G_found.shape==G_truth.shape==G_nan.shape)):
        print("Error shape for imputation")
    if mask is None:
        mask = np.logical_and(G_nan == 3, G_truth != 3)

    nb_nan = np.count_nonzero(mask)
    nb_diff = np.count_nonzero(G_found[mask] != G_truth[mask])

    try:
        return nb_diff / float(nb_nan)
    except ZeroDivisionError:
        return 0.0




import numpy as np
import itertools
import loter.tree as tree

from sklearn.cluster import KMeans

def init_data(data, inita, inith):
    data["A"] = inita(data["A"], data["G"]).astype(np.float32)
    data["H"] = inith(data["H"], data["G"])
    return data

def init_a_rand(A, G, randomstate=None,dtype=np.int):
    """Initialize the clusters randomly.

    Initialize an already allocated matrix A (ancestry,cluster,...)
    that is an numpy array with random values between 0 and 1.

    :param A:           Array to initialize
    :type A:            np array
    :param randomstate: RandomState of numpy
    :type randomstate:  np.random.Randomstate
    :param dtype:       Float or Integer values for the array
    :type dtype:        np.dtype
    :returns:           np.array -- Initialized A
    """

    if randomstate == None:
        randomstate = np.random.RandomState()

    (k,m) = A.shape
    if dtype is np.int:
        A = randomstate.randint(0,2,size=(k,m))
    else:
        A = randomstate.random_sample(size=(k,m), dtype=np.float)

    return A

def init_a_selection_g(A, G, randomstate=None, dtype=np.int):

    (k,m) = A.shape
    (n,_) = G.shape

    if randomstate == None:
        randomstate = np.random.RandomState()

    def snp_to_a(snp, dtype):
        if snp == 0:
            return 0
        elif snp == 2:
            return 1
        else:
            if dtype is np.int:
                return randomstate.randint(0,2)
            else:
                return randomstate.uniform()

    for j in range(m):
        selected_snps = randomstate.choice(n,k,replace=False)
        for i, pos_snps in enumerate(selected_snps):
            A[i,j] = snp_to_a(G[pos_snps,j], dtype)

    return A

def init_a_tree(A, G, randomstate=None, dtype=np.int):
    k, m = A.shape
    n, _ = G.shape

    if randomstate == None:
        randomstate = np.random.RandomState()

    rd_split = randomstate.randint(1, m-1)

    A = tree.mat_split(k, G, rd_split)

    return A

def init_a_kmeans(A, G):
    k, m = A.shape
    kmeans = KMeans(n_clusters=k).fit(G/2.0)

    A = np.clip(kmeans.cluster_centers_, 0, 1)
    return A

def init_a_deter(A, G):
    """Initialize the clusters deterministically.

    Initilalize an already allocated matrix A (ancestry,cluster,...)
    that is an numpy array with lines of ones and zeros.

    :param A:           Array to initialize
    :type A:            np array
    :returns:           np.array -- Initialized A
    """

    (k,m) = A.shape
    A[::2] = np.ones(m)

    return A

def init_h_deter(H,G):
    """Initialize the haplotypes deterministically.

    Initilalize an already allocated matrix H (haplotypes)
    that is an numpy array according to G.

    :param H:           Array to initialize
    :type H:            np array
    :param G:           Array of genotypes
    :type G:            np array
    :raises ValueError: Sizes are not compatible
    :returns:           np.array -- Initialized H
    """

    (n,m) = G.shape
    (hn,hm) = H.shape

    if (not hn == 2*n) or (not hm == m):
        raise ValueError("Sizes of H and G are not compatible.")

    def gi_to_hi(x,val):
        vlambd = np.vectorize(lambda xi: val if xi == 1 else xi/2)
        h_i = vlambd(x)

        return h_i

    H1 = np.apply_along_axis(gi_to_hi,1,G,val=0)
    H2 = np.apply_along_axis(gi_to_hi,1,G,val=1)

    H = np.zeros((hn,hm), dtype=np.uint8)
    H[::2] = H1
    H[1::2] = H2

    return H

def init_h_rand(H, G, randomstate=None):
    """Initialize the haplotypes randomly.

    Initilalize an already allocated matrix H (haplotypes)
    that is an numpy array such as according to G.

    :param H:           Array to initialize
    :type H:            np array
    :param G:           Array of genotypes
    :type G:            np array
    :raises ValueError: Sizes are not compatible
    :returns:           np.array -- Initialized H
    """

    if randomstate == None:
        randomstate = np.random.RandomState()

    (n,m) = G.shape
    (hn,hm) = H.shape

    if (not hn == 2*n) or (not hm == m):
        raise ValueError("Sizes of H and G are not compatible.")

    def gi_to_hi(x):
        vlambd = np.vectorize(lambda xi:
                              randomstate.randint(0,2) if xi == 1 else xi/2)
        h_i = vlambd(x)

        return h_i

    H1 = np.apply_along_axis(gi_to_hi,1,G)
    H2 = G - H1

    H = np.ascontiguousarray(np.zeros((hn,hm),dtype=np.uint8))
    H[::2] = H1
    H[1::2] = H2

    return H

def init_rand(data, param):
    return init_data(data, init_a_rand, init_h_rand)

def init_awithg_hrand(data, param):
    return init_data(data, init_a_selection_g, init_h_rand)

def init_atree_hrand(data, param):
    return init_data(data, init_a_tree, init_h_rand)

def init_deter(data, param):
    return init_data(data, init_a_deter, init_h_rand)

def init_kmeans(data, param):
    return init_data(data, init_a_kmeans, init_h_rand)

def create_data(G, param):
    n, m = G.shape
    k = param["nbclust"]
    data = {}
    data["G"] = G
    data["H"] = np.ascontiguousarray(np.zeros((2*n,m), dtype=np.uint8))
    data["S"] = np.ascontiguousarray(np.zeros((2*n,m), dtype=np.uint8))
    data["A"] = np.ascontiguousarray(np.zeros((k,m), dtype=np.float))

    return data

data_initializers = {
    "rand": init_rand,
    "deter": init_deter,
    "a with g, h rand": init_awithg_hrand,
    "a tree": init_awithg_hrand,
    "kmeans": init_kmeans
}



from __future__ import print_function
import inspect

import numpy as np
import scipy.stats as stats
import pandas as pd

import loter.pipeline as lt
import loter.initparam as initparam
import loter.initdata as initdata
import loter.opti as opti
import loter.estimatea as esta
import loter.estimateh as esth
import loter.graph as ests

##################################################################
#                                                                #
# Pipeline to optimize H (admixed haplotypes) and S (selection)  #
# given A (ancestral haplotypes) simultaneously                  #
#                                                                #
# Initialization:                                                #
# - init A with ancestral haplotypes                             #
#                                                                #
# Optimization:                                                  #
# - join optimisation of H and S                                 #
#                                                                #
# Computational Complexity:                                      #
# - n, the number of ancestral individuals                       #
# - m, the number of SNPs                                        #
# complexity -> O(n^2 * m)                                       #
#                                                                #
# Remarks:                                                       #
# In practice the pipeline is use as a phase corrector module    #
##################################################################

def init_fix_a(data, param):
    def init_a_fix(A, G):
        return param["A_in"]
    return initdata.init_data(data, init_a_fix, initdata.init_h_rand)

def opti_A_fix_join(data, param):
    data, param = ests.optimize_SHknn(data, param)
    data, param = esth.optimize_H_old(data, param)

    return data, param

fixa_pip_A_join = lt.Pipeline(
    initparam.param_initializers["classic_init"],
    init_fix_a,
    opti_A_fix_join,
)

##################################################################
#                                                                #
# Pipeline to optimize S (selection) given A                     #
# (ancestral haplotypes) and H (admixed haplotypes)              #
#                                                                #
# Initialization:                                                #
# - init A with ancestral haplotypes                             #
# - init H with admixed haplotypes                               #
#                                                                #
# Optimization:                                                  #
# - optimisation of S                                            #
#                                                                #
# Computational Complexity:                                      #
# - n, the number of ancestral individuals                       #
# - m, the number of SNPs                                        #
# complexity -> O(n * m)                                         #
#                                                                #
##################################################################

def init_fix_ah(data, param):
    def init_a_fix(A, G):
        return param["A_in"]
    def init_h_fix(H, G):
        return param["H_in"]
    return initdata.init_data(data, init_a_fix, init_h_fix)

def opti_AH_fix_knn(data, param):
    data["A"] = data["A"].astype(np.uint8)
    data["S"] = data["S"].astype(np.uint32)
    data, param = ests.optimize_Sknn(data, param)
    return data, param

def init_fix_ah_custom(data, param):
    def init_a_fix(A, G):
        return param["A_in"]
    def init_h_fix(H, G):
        return param["H_in"]

    data["anc_pop_id"] = param["anc_pop_id"]

    # trick 1 to pass lambda_ratio without modifying parameter class
    data["lambda_ratio"] = param["lambd_ratio"]

    return initdata.init_data(data, init_a_fix, init_h_fix)

def opti_AH_fix_knn_custom(data, param):

    data["A"] = data["A"].astype(np.uint8)
    data["S"] = data["S"].astype(np.uint32)
    data["anc_pop_id"] = data["anc_pop_id"].astype(np.uint32)

    # trick 2 to pass lambda_ratio without modifying parameter class
    param["lambda_ratio"] = data["lambda_ratio"]

    data, param = ests.optimize_Sknn_custom(data, param)

    return data, param

def init_fix_ah_custom2(data, param):
    def init_a_fix(A, G):
        return param["A_in"]
    def init_h_fix(H, G):
        return param["H_in"]

    data["anc_pop_id"] = param["anc_pop_id"]

    # trick 1 to pass lambda_ratio without modifying parameter class
    data["lambda_ratio"] = param["lambd_ratio"]
    data["mu"] = param["mu"]

    return initdata.init_data(data, init_a_fix, init_h_fix)

def opti_AH_fix_knn_custom2(data, param):

    data["A"] = data["A"].astype(np.uint8)
    data["S"] = data["S"].astype(np.uint32)
    data["anc_pop_id"] = data["anc_pop_id"].astype(np.uint32)

    # trick 2 to pass lambda_ratio without modifying parameter class
    param["lambda_ratio"] = data["lambda_ratio"]
    param["mu"] = data["mu"]

    data, param = ests.optimize_Sknn_custom2(data, param)

    return data, param

fixa_pip_AH_knn = lt.Pipeline(
    initparam.param_initializers["classic_init"],
    init_fix_ah,
    opti_AH_fix_knn
)

fixa_pip_AH_knn_custom = lt.Pipeline(
    initparam.param_initializers["classic_init"],
    init_fix_ah_custom,
    opti_AH_fix_knn_custom
)

fixa_pip_AH_knn_custom2 = lt.Pipeline(
    initparam.param_initializers["classic_init"],
    init_fix_ah_custom2,
    opti_AH_fix_knn_custom2
)

def learn_Sknn(pop, A_in, H_in, weights, anc_pop_id=None,
               penalty=40, lambd_ratio=1, mu=None, num_threads=10):
    G_pop = pop["G"]
    H_pop = H_in

    if mu is not None and lambd_ratio is not None:
        l_res_mix = fixa_pip_AH_knn_custom2(G_pop,
                                           nb_iter=1, nbclust=len(A_in), penalty=penalty,
                                           num_threads=num_threads,
                                           weights=weights,
                                           A_in=A_in,
                                           H_in=H_pop,
                                           anc_pop_id=anc_pop_id,
                                           lambd_ratio=lambd_ratio,
                                           mu=mu
        )
    elif lambd_ratio is not None:
        l_res_mix = fixa_pip_AH_knn_custom(G_pop,
                                           nb_iter=1, nbclust=len(A_in), penalty=penalty,
                                           num_threads=num_threads,
                                           weights=weights,
                                           A_in=A_in,
                                           H_in=H_pop,
                                           anc_pop_id=anc_pop_id,
                                           lambd_ratio=lambd_ratio
        )
    else:
        l_res_mix = fixa_pip_AH_knn(G_pop,
                                    nb_iter=1, nbclust=len(A_in), penalty=penalty,
                                    num_threads=num_threads,
                                    weights=weights,
                                    A_in=A_in,
                                    H_in=H_pop
        )

    return l_res_mix[0]["S"]

def learn_S_join(pop, A_in, penalty=40, small_penalty=0, num_threads=10):
    G_pop = pop["G"]

    l_res_mix = fixa_pip_A_join(G_pop,
                                nb_iter=1, nbclust=len(A_in), penalty=penalty,
                                A_in=A_in,
                                small_penalty=small_penalty,
                                num_threads=num_threads
    )

    return l_res_mix[0]["S"], l_res_mix[0]["H"]

def get_items(dict_object):
    """
    Compatible Python 2 et 3 get item for dictionnary
    """
    for key in dict_object:
        yield key, dict_object[key]

def clusters_to_list_pop(S, l_k):
    """
    From a selection matrix S, compute the origin of each SNP.

    input:
    S -- matrix where we are copying
    l_k -- populations sizes
    """
    res = np.copy(S)

    a = np.repeat(np.arange(len(l_k)), l_k)
    b = np.arange(sum(l_k))
    d = {k: v for v, k in zip(a, b)}
    for k, v in get_items(d): res[S==k] = v
    return res

def locanc_g_knn(l_h, g_adm, penalty=40, small_penalty=0, num_threads=10):
    A_in = np.ascontiguousarray(np.vstack(l_h))
    S_adm, H = learn_S_join({"G": g_adm}, A_in, penalty, small_penalty, num_threads)
    result = clusters_to_list_pop(S_adm, [len(A) for A in l_h])

    return result, S_adm, H

def locanc_h_knn(l_h, h_adm, penalty=40, lambd_ratio=1, mu=None, num_threads=10):
    A_in = np.ascontiguousarray(np.vstack(l_h))
    g_adm = h_adm[::2] + h_adm[1::2]
    n, m = h_adm.shape
    weights = np.ones(m)

    anc_pop_size = [len(A) for A in l_h]
    anc_pop_id = np.repeat(np.arange(len(anc_pop_size)), anc_pop_size)

    S_adm = learn_Sknn({"G": g_adm, "H": h_adm}, A_in, h_adm,
                       weights, anc_pop_id, penalty, lambd_ratio, mu, num_threads)
    result = clusters_to_list_pop(S_adm, anc_pop_size)

    return result, S_adm

def update_counts(counts, arr, k=2):
    for p in range(k):
        counts[p,:,:][arr == p] += 1
    return counts

def mode(counts):
    argmax = np.argmax(counts, axis=0)
    return argmax, argmax.choose(counts)

def encode_haplo(H):
    H1, H2 = H[::2], H[1::2]
    return ((np.maximum(H1, H2) * (np.maximum(H1, H2) + 1)) / 2) + np.minimum(H1, H2)

def loter_multiple_pops(l_H, h_adm, lambd, lambd_ratio=None, mu=None,
                        num_threads=10, verbosity=False):
    """
    Local ancestry inference.

    Args:
        l_H (list of np.array): list of matrices of ancestral individual
            haplotypes from different populations (one matrix of dimension
            `2n_pop x n_snp` per ancestral population, where `n_pop` is the
            size of the population and `n_snp` the number of SNPs).
        h_adm (np.array): matrix of admixed individual haplotypes of
            dimension `2n x n_snp`.
        lambd (float): positive real value, penalty parameter. Default
            value is 1.
        lambda_ratio (float): positive real value. Default is None.
            If not None, two different lambdas (`lambda_inter` and
            `lambda_intra`) are used in the inference algorithm, to allow
            different penalization for intra and inter ancetral population
            switch. In this case, `lambda_inter` is given by `range_lambda`
            and `lambda_intra = lambda_inter * lambda_ratio`.
        mu (float): positive real value. Default is None. If not None, the
            local ancestry inference algorithm includes a phase error
            correction penalization term where `mu` is the penalty parameter.
        threshold (float): real value in [0,1], threshold for the phase
            correction module. Default value is 0.9.
        rate_vote (float): real value in [0,1], threshold for vote
            in bagging. Default value is 0.5.
        nb_bagging (int): positive integer value, number of sub-sampling in
            bagging. Default value is 20.
        nb_threads (int): positive integer, number of threads to use. If =1
            (default value), a single thread is use. If >1, on OpenMP
            compatible system, computations are parallelized.
        verbosity (bool): enable/disable versbosity. Default is False.

    Returns:
        Matrix (np.array) of infered ancestral population id for each SNPs
        and each admixed haplotype (dimension `2n x n_snp`).
    """
    if verbosity:
        print("# Run simple local ancestry inference")
        print("# Input:")
        frame = inspect.currentframe()
        args, _, _, values = inspect.getargvalues(frame)
        for i in args: print("## {} = {}".format(i, values[i]))
    res_loter, _ = locanc_h_knn([h.astype(np.uint8) for h in l_H],
                               h_adm.astype(np.uint8),
                               lambd, lambd_ratio, mu, num_threads)
    return res_loter

def boostrap_loter_multiple_pops(l_H, h_adm, lambd, lambd_ratio, mu,
                                 counts, nbrun=20, num_threads=10,
                                 verbosity=False):

    if verbosity:
        print("# Bootstrap local ancestry inference")
        print("# Input:")
        frame = inspect.currentframe()
        args, _, _, values = inspect.getargvalues(frame)
        for i in args: print("## {} = {}".format(i, values[i]))

    def shuffle(H):
        n, m = H.shape
        return H[np.random.randint(n, size=n), :]

    if nbrun > 1:
        if verbosity: print("## Multiple run")
        for i in range(nbrun):
            shuffled_H = [shuffle(h) for h in l_H]
            counts = update_counts(counts,
                                   loter_multiple_pops(shuffled_H,
                                                       h_adm,
                                                       lambd,
                                                       lambd_ratio,
                                                       mu,
                                                       num_threads,
                                                       verbosity=False),
                                   len(l_H)
            )
    else:
        if verbosity: print("## Single run")
        counts = update_counts(counts,
                               loter_multiple_pops(l_H,
                                                   h_adm,
                                                   lambd,
                                                   lambd_ratio,
                                                   mu,
                                                   num_threads,
                                                   verbosity=False),
                               len(l_H)
        )

    return counts

def loter_local_ancestry(l_H, h_adm, range_lambda=np.arange(1.5, 5.5, 0.5),
                         lambda_ratio=None, mu=None,
                         rate_vote=0.5, nb_bagging=20,
                         num_threads=10, verbosity=False):
    """
    Local ancestry inference with bagging.

    Args:
        l_H (list of np.array): list of matrices of ancestral individual
            haplotypes from different populations (one matrix of dimension
            `2n_pop x n_snp` per ancestral population, where `n_pop` is the
            size of the population and `n_snp` the number of SNPs).
        h_adm (np.array): matrix of admixed individual haplotypes of
            dimension `2n x n_snp`.
        range_lambda (np.array): vector of (positive real) candidates values
            for the penalty parameter. Default value is
            {1.5,2,2.5,3,3.5,4,4.5,5,5.5}.
        lambda_ratio (float or np.array): positive single value or vector of
            positive real value. Default is None. If not None,
            two different lambdas (`lambda_inter`
            and `lambda_intra`) are used in the inference algorithm,
            to allow different penalization for intra and inter ancetral
            population switch. In this case, `lambda_inter` is given by
            `range_lambda` and `lambda_intra = lambda_inter * lambda_ratio`.
            If `lambda_ratio` is a vector, the different values will be tested
            in the bagging algorithm.
        mu (float or np.array): single positive real value or vector of
            positive real values. Default is None. If not None, the
            local ancestry inference algorithm includes a phase error
            correction penalization term where `mu` is the penalty parameter.
            If `mu` is a vector, the different values will be tested
            in the bagging algorithm.
        rate_vote (float): real value in [0,1], threshold for vote
            in bagging. Default value is 0.5.
        nb_bagging (int): positive integer value, number of sub-sampling in
            bagging. Default value is 20.
        nb_threads (int): positive integer, number of threads to use. If =1
            (default value), a single thread is use. If >1, on OpenMP
            compatible system, computations are parallelized.
        verbosity (bool): enable/disable versbosity. Default is False.

    Returns:
        r (np.array): matrix of infered genotype for each SNPs and each
            admixed haplotype (dimension `n x n_snp`).
        res_loter (tuple): result from boostrap, tuple of two matrices
            (np.array of dimension `2n x n_snp`), first is the matrix of
            ancestral population that was chosen by most of
            boostrap repetitions for each admixed haplotype and each SNP,
            second is the corresponding number of bootstrap repetitions that
            chose it.
    """
    if verbosity:
        print("# Run local ancestry inference (with bagging)")
        print("# Input:")
        frame = inspect.currentframe()
        args, _, _, values = inspect.getargvalues(frame)
        for i in args: print("## {} = {}".format(i, values[i]))
    input_loter = (l_H, h_adm)
    n, m = h_adm.shape
    counts = np.zeros((len(l_H), n, m))
    if not isinstance(range_lambda, np.ndarray):
        if isinstance(range_lambda, list) or isinstance(range_lambda, tuple):
            range_lambda = np.array(range_lambda)
        else:
            range_lambda = np.array([range_lambda])
    if not isinstance(lambda_ratio, np.ndarray):
        if isinstance(lambda_ratio, list) or isinstance(lambda_ratio, tuple):
            lambda_ratio = np.array(lambda_ratio)
        else:
            lambda_ratio = np.array([lambda_ratio])
    if not isinstance(mu, np.ndarray):
        if isinstance(mu, list) or isinstance(mu, tuple):
            mu = np.array(mu)
        else:
            mu = np.array([mu])
    for l, lr, m in np.dstack(np.meshgrid(range_lambda, lambda_ratio, mu)).reshape(-1, 3):
        res_boostrap = boostrap_loter_multiple_pops(*input_loter, lambd=l,
                                                    lambd_ratio=lr, mu=m,
                                                    counts=counts, nbrun=nb_bagging,
                                                    num_threads=num_threads,
                                                    verbosity=False)
    res_loter = mode(counts)
    r = vote_and_impute(res_loter, rate_vote)
    return r, res_loter

def diploid_sim(cluster_found, cluster_truth):
    (n,m) = cluster_found.shape
    return np.count_nonzero(cluster_found == cluster_truth) / float(n*m)

def find_lambda(s_in, threshold = 0.90, min_lambda = 1,
                max_lambda = 500, num_threads=10):
    n, m = s_in.shape
    if max_lambda - min_lambda <= 1:
        return locanc_g_knn([np.zeros((1,m)), np.ones((1,m))],
                            s_in, min_lambda, min_lambda, num_threads)
    else:
        mean = (max_lambda - min_lambda) / 2 + min_lambda
        r_g, s_g, h_g = locanc_g_knn([np.zeros((1,m)), np.ones((1,m))],
                                     s_in, mean, mean, num_threads)
        sim = diploid_sim(r_g[::2] + r_g[1::2], s_in)
        if sim > threshold:
            return find_lambda(s_in, threshold, min_lambda = (max_lambda - min_lambda) / 2 + min_lambda,
                               max_lambda = max_lambda, num_threads=num_threads)
        else:
            return find_lambda(s_in, threshold, min_lambda = min_lambda,
                               max_lambda = max_lambda  - ((max_lambda - min_lambda) / 2),
                               num_threads=num_threads)

def vote_and_impute(s, percent_threshold=0.5):
    def select_val(s, percent_threshold):
        max_s, min_s = np.max(s[1]), np.min(s[1])
        threshold = percent_threshold*(max_s - min_s) + min_s
        select = np.logical_and(s[1][::2] >= threshold,
                                s[1][1::2] >= threshold)
        arr = encode_haplo(s[0])
        arr[np.logical_not(select)] = 255
        return arr

    arr = select_val(s, percent_threshold)

    n, m = arr.shape
    res = np.copy(arr)

    for i in range(n):
        serie = pd.Series(arr[i])
        serie.loc[serie == 255] = np.nan
        try:
            res[i] = serie.dropna().reindex(range(m), method='nearest').values
        except:
            res[i] = arr[i]

    return res

def loter_smooth(l_H, h_adm, range_lambda=np.arange(1.5, 5.5, 0.5),
                 lambda_ratio=None, threshold=0.90, rate_vote=0.5,
                 nb_bagging=20, num_threads=10,
                 verbosity=False):
    """
    Local ancestry inference with bagging and phase correction.

    Args:
        l_H (list of np.array): list of matrices of ancestral individual
            haplotypes from different populations (one matrix of dimension
            `2n_pop x n_snp` per ancestral population, where `n_pop` is the
            size of the population and `n_snp` the number of SNPs).
        h_adm (np.array): matrix of admixed individual haplotypes of
            dimension `2n x n_snp`.
        range_lambda (np.array): vector of (positive real) candidates values
            for the penalty parameter. Default value is
            {1.5,2,2.5,3,3.5,4,4.5,5,5.5}.
        lambda_ratio (float or np.array): single value or vector of real value.
            Default is None. If not None, two different lambdas (`lambda_inter`
            and `lambda_intra`) are used in the inference algorithm,
            to allow different penalization for intra and inter ancetral
            population switch. In this case, `lambda_inter` is given by
            `range_lambda` and `lambda_intra = lambda_inter * lambda_ratio`.
            If `lambda_ratio` is a vector, the different values will be tested
            in the bagging algorithm.
        threshold (float): real value in [0,1], threshold for the phase
            correction module. Default value is 0.9.
        rate_vote (float): real value in [0,1], threshold for vote
            in bagging. Default value is 0.5.
        nb_bagging (int): positive integer value, number of sub-sampling in
            bagging. Default value is 20.
        nb_threads (int): positive integer, number of threads to use. If =1
            (default value), a single thread is use. If >1, on OpenMP
            compatible system, computations are parallelized.
        verbosity (bool): enable/disable versbosity. Default is False.

    Returns:
        Matrix (np.array) of infered ancestral population id for each SNPs
        and each admixed haplotype (dimension `2n x n_snp`).
    """
    if verbosity:
        print("# Running local ancestry inference")
        print("# Input:")
        frame = inspect.currentframe()
        args, _, _, values = inspect.getargvalues(frame)
        for i in args: print("## {} = {}".format(i, values[i]))

    if verbosity: print("## Step 1: local ancestry inference (with bagging)")
    res_impute, res_raw = loter_local_ancestry(
                                l_H=l_H, h_adm=h_adm, range_lambda=range_lambda,
                                lambda_ratio=lambda_ratio, mu=None,
                                rate_vote=rate_vote, nb_bagging=nb_bagging,
                                num_threads=num_threads,
                                verbosity=False)
    result = np.copy(res_impute)
    result_hap = []
    if verbosity: print("## Step 2: phase correction")
    for i in range(len(res_impute)):
        arr_input = np.ascontiguousarray(np.array([res_impute[i]])).astype(np.uint8)
        r, _, _ = find_lambda(arr_input, threshold=threshold, num_threads=num_threads)
        result_hap.append(r)
        result[i] = r[::2] + r[1::2]

    return np.vstack(result_hap)



from __future__ import division

import unittest
import numpy as np

import loter.estimatea as estimatea
import tests.generate_input as gen

class EstimateATest(unittest.TestCase):

    def setUp(self):
        self.data, self.param = gen.generate_simple_data_param()

    def test_estimatea(self):
        esta = estimatea.EstimateA()
        esta.run(self.data, self.param)

        res = np.array([[1,1,0,0],
                        [2/3,2/3,2/3,0],
                        [1,1,0,0]])

        self.assertTrue(np.allclose(res,self.data["A"]))

    def test_estimatea_nan(self):
        esta = estimatea.EstimateA()
        self.data["G"][0,0] = 3
        self.data["G"][1,2] = 3
        self.data["G"][2,1] = 3
        esta.run(self.data, self.param)

        res = np.array([[0.5,1,0,0],
                        [0.5,2/3,1,0],
                        [1,0.5,0,0]])

        self.assertTrue(np.allclose(res,self.data["A"]))



def build_init_param(penalty=2.0, nbclust=10, w_h=100, nb_iter=10):
    def init_param(G, **kwargs):
        param = {"penalty": penalty,
                "nbclust": nbclust,
                "w_h": w_h,
                "nb_iter": nb_iter}

        for key, value in list(kwargs.items()):
            param[key] = value

        return param

    return init_param

param_initializers = {
    "classic_init": build_init_param()
}



import ctypes as C
import numpy as np
import random
from collections import namedtuple

import loter.errorhandler as errorhandler
import loter.datastruct.parameter as parameter
import loter.utils as utils
from loter.find_lib import _LIB

@errorhandler.eh_fn
def EstimateHErrorHandlerFn(error_message, user_data):
    """Callback function for C api in errorhandler.c

    """
    print("Error Estimation of H")
    print(error_message)

class EstimateHException(Exception):
    """Exception raised for errors in EstimateA calls

    Attributes:
        value -- input in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, value, message):
        self.value = value
        self.message = message

def _init_estimateh():

    # Constructor selectswitch
    _LIB.selectswitch_create.argtypes = [
        C.POINTER(errorhandler.ErrorHandler)
    ]
    _LIB.selectswitch_create.restype = C.c_void_p
    # Destructor selectswitch
    _LIB.selectswitch_destroy.argtypes = [
        C.c_void_p,
        C.POINTER(errorhandler.ErrorHandler)
    ]
    _LIB.selectswitch_destroy.restype = C.c_void_p

     # Constructor selectswitchdeter
    _LIB.selectswitchdeter_create.argtypes = [
        C.POINTER(errorhandler.ErrorHandler)
    ]
    _LIB.selectswitchdeter_create.restype = C.c_void_p
    # Destructor selectswitchdeter
    _LIB.selectswitchdeter_destroy.argtypes = [
        C.c_void_p,
        C.POINTER(errorhandler.ErrorHandler)
    ]
    _LIB.selectswitchdeter_destroy.restype = C.c_void_p

     # Constructor selectswitchprob
    _LIB.selectswitchprob_create.argtypes = [
        C.POINTER(errorhandler.ErrorHandler)
    ]
    _LIB.selectswitchprob_create.restype = C.c_void_p
    # Destructor selectswitchprob
    _LIB.selectswitchprob_destroy.argtypes = [
        C.c_void_p,
        C.POINTER(errorhandler.ErrorHandler)
    ]
    _LIB.selectswitchprob_destroy.restype = C.c_void_p

   # Constructor treatnan
    _LIB.treatnan_create.argtypes = [
        C.POINTER(errorhandler.ErrorHandler)
    ]
    _LIB.treatnan_create.restype = C.c_void_p
    # Destructor treatnan
    _LIB.treatnan_destroy.argtypes = [
        C.c_void_p,
        C.POINTER(errorhandler.ErrorHandler)
    ]
    _LIB.treatnan_destroy.restype = C.c_void_p


    # Constructor
    _LIB.estimateh_create.argtypes = [
        C.c_void_p,
        C.c_void_p,
        C.POINTER(errorhandler.ErrorHandler)
    ]
    _LIB.estimateh_create.restype = C.c_void_p

    # Destructor
    _LIB.estimateh_destroy.argtypes = [
        C.c_void_p,
        C.POINTER(errorhandler.ErrorHandler)
    ]
    _LIB.estimateh_destroy.restype = None

    # Getter Penalty
    _LIB.estimateh_run.argtypes = [
        C.c_void_p,
        np.ctypeslib.ndpointer(dtype = np.uint8,
                               ndim=2,
                               flags='C_CONTIGUOUS'),
        np.ctypeslib.ndpointer(dtype = np.uint8,
                               ndim=2,
                               flags='C_CONTIGUOUS'),
        np.ctypeslib.ndpointer(dtype = np.float32,
                               ndim=2,
                               flags='C_CONTIGUOUS'),
        np.ctypeslib.ndpointer(dtype = np.uint8,
                               ndim=2,
                               flags='C_CONTIGUOUS'),
        C.c_int,
        C.c_int,
        C.c_void_p,
        C.POINTER(errorhandler.ErrorHandler)
    ]
    _LIB.estimateh_run.restype = None

class SelectSwitch(object):
    def __init__(self):
        self._EH = errorhandler.ErrorHandler(EstimateHErrorHandlerFn, None)
        self.obj = _LIB.selectswitch_create(C.byref(self._EH))

    def __del__(self):
        _LIB.selectswitch_destroy(self.obj,
                                  C.byref(self._EH))

class SelectSwitchDeter(SelectSwitch):
    def __init__(self):
        self._EH = errorhandler.ErrorHandler(EstimateHErrorHandlerFn, None)
        self.obj = _LIB.selectswitchdeter_create(C.byref(self._EH))

    def __del__(self):
        _LIB.selectswitchdeter_destroy(self.obj,
                                       C.byref(self._EH))

class SelectSwitchProb(SelectSwitch):
    def __init__(self):
        self._EH = errorhandler.ErrorHandler(EstimateHErrorHandlerFn, None)
        self.obj = _LIB.selectswitchprob_create(C.byref(self._EH))

    def __del__(self):
        _LIB.selectswitchprob_destroy(self.obj,
                                      C.byref(self._EH))

class TreatNaN(object):
    def __init__(self):
        self._EH = errorhandler.ErrorHandler(EstimateHErrorHandlerFn, None)
        self.obj = _LIB.treatnan_create(C.byref(self._EH))

    def __del__(self):
        _LIB.treatnan_destroy(self.obj,
                              C.byref(self._EH))

wrapped = namedtuple("wrapped", "obj")
class EstimateH(object):

    def __init__(self, sswitch = wrapped(obj = None), ttnan = wrapped(obj = None)):
        self._EH = errorhandler.ErrorHandler(EstimateHErrorHandlerFn, None)
        self.obj = _LIB.estimateh_create(sswitch.obj,
                                         ttnan.obj,
                                         C.byref(self._EH))

    def run(self, data, param):

        param_opti = parameter.ParameterOptimization(param["penalty"],
                                                     param["nbclust"],
                                                     param["w_h"],
                                                     param["nb_iter"])

        G, H, A, S = data["G"], data["H"], data["A"], data["S"]
        n, m = G.shape

        _LIB.estimateh_run(self.obj,
                           G,
                           H,
                           A,
                           S,
                           n,
                           m,
                           param_opti.obj,
                           C.byref(self._EH))

    def __del__(self):
        _LIB.estimateh_destroy(self.obj, C.byref(self._EH))

def optimize_H(data, param):
    esth = EstimateH()
    esth.run(data, param)
    return (data, param)

def optimize_H_old(data, param):
    sswitch = SelectSwitchDeter()
    esth = EstimateH(sswitch)
    esth.run(data, param)
    return (data, param)

def optimize_H_prob(data, param):
    sswitch = SelectSwitchProb()
    esth = EstimateH(sswitch)
    esth.run(data, param)
    return (data, param)

def optimize_H_old_fast(data, param):
    random.seed()
    def H_opti(data, param):
        G = data["G"]
        (n, m) = G.shape
        A = data["A"]
        H1, H2 = data["H"][::2], data["H"][1::2]
        S1, S2 = data["S"][::2], data["S"][1::2]
        A1, A2 = utils.build_SA(S1, A), utils.build_SA(S2, A)

        mask_a1_over_a2 = np.logical_and(G == 1, A1 >= A2)
        mask_a1_under_a2 = np.logical_and(G == 1, A1 < A2)
        H1[mask_a1_over_a2] = 1
        H2[mask_a1_over_a2] = 0
        H1[mask_a1_under_a2] = 0
        H2[mask_a1_under_a2] = 1

        H1[G == 3] = np.around(A1[G == 3])
        H2[G == 3] = np.around(A2[G == 3])
        data["H"][::2] = H1
        data["H"][1::2] = H2
        return (data, param)

    return H_opti(data, param)

def optimize_H_sample_nan_prob(data, param):
    random.seed()
    def H_opti(data, param):
        G = data["G"]
        (n, m) = G.shape
        A = data["A"]
        H1, H2 = data["H"][::2], data["H"][1::2]
        S1, S2 = data["S"][::2], data["S"][1::2]
        for i in range(n):
            for j in range(m):
                if G[i,j] == 1:
                    a1, a2 =  A[S1[i,j],j], A[S2[i,j],j]
                    norm_prob = a1 + a2
                    if norm_prob == 0:
                        prob_a = 0.5
                    else:
                        prob_a = float(a1) / norm_prob
                    if self.rng.uniform(0,1) <= prob_a:
                        H1[i,j] = 1
                        H2[i,j] = 0
                    else:
                        H1[i,j] = 0
                        H2[i,j] = 1
                elif G[i,j] == 3:
                    H1[i,j] = np.around(A[S1[i,j],j])
                    H2[i,j] = np.around(A[S2[i,j],j])

        data["H"][::2] = H1
        data["H"][1::2] = H2
        return (data, param)
    return H_opti(data, param)

#initialize
_init_estimateh()

optih = {
    "opti_cpp": optimize_H,
    "opti_prob": optimize_H_prob,
    "opti py old": optimize_H_old,
    "opti py old fast": optimize_H_old_fast,
    "opti py fast": optimize_H_sample_nan_prob
}



from __future__ import division

import unittest
import numpy as np

import loter.graph as estimates
import tests.generate_input as gen

class EstimateSTest(unittest.TestCase):

    def setUp(self):
        self.G = np.array([[2,2,0,0],
                           [1,1,0,0],
                           [1,1,1,1]], dtype=np.uint8)
        self.A = np.array([[1,1,0,0],
                           [0,0,1,1]], dtype=np.float32)
        self.H = np.array([[1,1,0,0],
                           [1,1,0,0],
                           [1,0,0,0],
                           [0,1,0,0],
                           [1,1,1,1],
                           [0,0,0,0]], dtype=np.uint8)
        self.S = np.zeros((6,4), dtype=np.uint8)
        self.data = {
            "G": self.G,
            "H": self.H,
            "A": self.A,
            "S": self.S,
        }
        self.param = {
            "penalty": 0.0,
            "nbclust": 2,
            "nb_iter": 10,
            "w_h": 1
        }

    def test_estimates(self):
        ests = estimates.EstimateS()
        ests.run(self.data, self.param)
        res = np.array([[ 0,  0,  0,  0],
                        [ 0,  0,  0,  0],
                        [ 0,  1,  0,  0],
                        [ 1,  0,  0,  0],
                        [ 0,  0,  1,  1],
                        [ 1,  1,  0,  0]])

        self.assertTrue(np.allclose(res, self.data["S"], atol=0.01))

    def test_estimates_penalty(self):
        self.data["A"] = np.array([[1,1,1,0],
                                   [0,0,0,1]], dtype=np.float32)
        self.param["penalty"] = 5.0

        ests = estimates.EstimateS()
        ests.run(self.data, self.param)
        res = np.array([[ 0,  0,  0,  0],
                        [ 0,  0,  0,  0],
                        [ 0,  0,  0,  0],
                        [ 0,  0,  0,  0],
                        [ 0,  0,  0,  0],
                        [ 1,  1,  1,  1]])
        self.assertTrue(np.allclose(res, self.data["S"], atol=0.01))

    def test_estimates_nan(self):
        self.data["G"][0,0] = 3
        self.data["G"][1,1] = 3
        self.data["G"][2,2] = 3
        self.param["penalty"] = 0.0


        ests = estimates.EstimateS()
        ests.run(self.data, self.param)
        res = np.array([[ 0,  0,  0,  0],
                        [ 0,  0,  0,  0],
                        [ 0,  0,  0,  0],
                        [ 1,  1,  0,  0],
                        [ 0,  0,  0,  1],
                        [ 1,  1,  1,  0]])

        self.assertTrue(np.allclose(res, self.data["S"], atol=0.01))

class EstimateSHTest(unittest.TestCase):

    def setUp(self):
        self.G = np.array([[2,2,0,0],
                           [1,1,0,0],
                           [1,1,1,1]], dtype=np.uint8)
        self.A = np.array([[1,1,0,0],
                           [0,0,1,1]], dtype=np.float32)
        self.H = np.array([[1,1,0,0],
                           [1,1,0,0],
                           [1,0,0,0],
                           [0,1,0,0],
                           [1,1,1,1],
                           [0,0,0,0]], dtype=np.uint8)
        self.S = np.zeros((6,4), dtype=np.uint8)
        self.data = {
            "G": self.G,
            "H": self.H,
            "A": self.A,
            "S": self.S,
        }
        self.param = {
            "penalty": 0.0,
            "nbclust": 2,
            "nb_iter": 10,
            "w_h": 1
        }

    def test_estimatesh_penalty(self):
        self.data["A"] = np.array([[1,1,1,0],
                                   [0,0,0,1]], dtype=np.float32)
        self.param["penalty"] = 15.0

        ests = estimates.EstimateSH()
        ests.run(self.data, self.param)
        res = np.array([[ 0,  0,  0,  0],
                        [ 0,  0,  0,  0],
                       	[ 1,  1,  1,  1],
                        [ 0,  0,  0,  0],
                        [ 1,  1,  1,  1],
                        [ 0,  0,  0,  0]])

        self.assertTrue(np.allclose(res, self.data["S"], atol=0.01))



from __future__ import division

import unittest
import numpy as np

import loter.estimateh as estimateh
import tests.generate_input as gen

class EstimateHTest(unittest.TestCase):

    def setUp(self):
        self.data, self.param = gen.generate_simple_data_param()

    def test_estimateh(self):
        esth = estimateh.EstimateH()
        l_H = []
        for i in range(10000):
            esth.run(self.data, self.param)
            l_H.append(self.data["H"].astype(np.float32))

        mean_H = np.average(np.dstack(l_H), axis=2)

        res = np.array([[1,1,1,0],
                        [1,1,0,0],
                        [0.5,0.5,0.5,0],
                        [0.5,0.5,0.5,0],
                        [1,1,0,0],
                        [1,1,0,0]
        ])

        self.assertTrue(np.allclose(res, mean_H, atol=0.01))

    def test_estimateh_nan(self):
        self.data["G"][0,0] = 3
        self.data["G"][1,2] = 3
        self.data["G"][2,1] = 3
        esth = estimateh.EstimateH()
        l_H = []
        for i in range(10000):
            esth.run(self.data, self.param)
            l_H.append(self.data["H"].astype(np.float32))

        mean_H = np.average(np.dstack(l_H), axis=2)

        res = np.array([[0,1,1,0],
                        [1,1,0,0],
                        [0.5,0.5,0,0],
                        [0.5,0.5,0,0],
                        [1,1,0,0],
                        [1,1,0,0]
        ])

        self.assertTrue(np.allclose(res, mean_H, atol=0.01))



import ctypes as C
import numpy as np

import loter.errorhandler as errorhandler
import loter.datastruct.parameter as parameter
from loter.find_lib import _LIB

from scipy.optimize import linear_sum_assignment as linear_assignment  # Munkres Algorithm
from collections import Counter

@errorhandler.eh_fn
def EstimateAErrorHandlerFn(error_message, user_data):
    """Callback function for C api in errorhandler.c

    """
    print("Error Estimation of A")
    print(error_message)

class EstimateAException(Exception):
    """Exception raised for errors in EstimateA calls

    Attributes:
        value -- input in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, value, message):
        self.value = value
        self.message = message

def _init_estimatea():

    # Constructor
    _LIB.estimatea_create.argtypes = [
        C.POINTER(errorhandler.ErrorHandler)
    ]
    _LIB.estimatea_create.restype = C.c_void_p

    # Destructor
    _LIB.estimatea_destroy.argtypes = [
        C.c_void_p,
        C.POINTER(errorhandler.ErrorHandler)
    ]
    _LIB.estimatea_destroy.restype = None

    # Getter Penalty
    _LIB.estimatea_run.argtypes = [
        C.c_void_p,
        np.ctypeslib.ndpointer(dtype = np.uint8,
                               ndim=2,
                               flags='C_CONTIGUOUS'),
        np.ctypeslib.ndpointer(dtype = np.uint8,
                               ndim=2,
                               flags='C_CONTIGUOUS'),
        np.ctypeslib.ndpointer(dtype = np.float32,
                               ndim=2,
                               flags='C_CONTIGUOUS'),
        np.ctypeslib.ndpointer(dtype = np.uint8,
                               ndim=2,
                               flags='C_CONTIGUOUS'),
        C.c_int,
        C.c_int,
        C.c_void_p,
        C.POINTER(errorhandler.ErrorHandler)
    ]
    _LIB.estimatea_run.restype = None

    # Constructor
    _LIB.estimateagrad_create.argtypes = [
        C.POINTER(errorhandler.ErrorHandler)
    ]
    _LIB.estimateagrad_create.restype = C.c_void_p

    # Destructor
    _LIB.estimateagrad_destroy.argtypes = [
        C.c_void_p,
        C.POINTER(errorhandler.ErrorHandler)
    ]
    _LIB.estimateagrad_destroy.restype = None

    # Getter Penalty
    _LIB.estimateagrad_run.argtypes = [
        C.c_void_p,
        np.ctypeslib.ndpointer(dtype = np.uint8,
                               ndim=2,
                               flags='C_CONTIGUOUS'),
        np.ctypeslib.ndpointer(dtype = np.uint8,
                               ndim=2,
                               flags='C_CONTIGUOUS'),
        np.ctypeslib.ndpointer(dtype = np.float32,
                               ndim=2,
                               flags='C_CONTIGUOUS'),
        np.ctypeslib.ndpointer(dtype = np.uint8,
                               ndim=2,
                               flags='C_CONTIGUOUS'),
        C.c_int,
        C.c_int,
        C.c_void_p,
        C.POINTER(errorhandler.ErrorHandler)
    ]
    _LIB.estimateagrad_run.restype = None

class EstimateA(object):

    def __init__(self):
        self._EH = errorhandler.ErrorHandler(EstimateAErrorHandlerFn, None)
        self.obj = _LIB.estimatea_create(C.byref(self._EH))

    def run(self, data, param):

        param_opti = parameter.ParameterOptimization(param["penalty"],
                                                     param["nbclust"],
                                                     param["w_h"],
                                                     param["nb_iter"])

        G, H, A, S = data["G"], data["H"], data["A"], data["S"]
        n, m = G.shape

        _LIB.estimatea_run(self.obj,
                           G,
                           H,
                           A,
                           S,
                           n,
                           m,
                           param_opti.obj,
                           C.byref(self._EH))

    def __del__(self):
        _LIB.estimatea_destroy(self.obj, C.byref(self._EH))

class EstimateAGrad(object):

    def __init__(self):
        self._EH = errorhandler.ErrorHandler(EstimateAErrorHandlerFn, None)
        self.obj = _LIB.estimateagrad_create(C.byref(self._EH))

    def run(self, data, param):

        param_opti = parameter.ParameterOptimization(param["penalty"],
                                                     param["nbclust"],
                                                     param["w_h"],
                                                     param["nb_iter"])

        G, H, A, S = data["G"], data["H"], data["A"], data["S"]
        n, m = G.shape

        _LIB.estimateagrad_run(self.obj,
                               G,
                               H,
                               A,
                               S,
                               n,
                               m,
                               param_opti.obj,
                               C.byref(self._EH))

    def __del__(self):
        _LIB.estimateagrad_destroy(self.obj, C.byref(self._EH))

def optimize_A(data, param):
    esta = EstimateA()
    esta.run(data, param)
    return (data, param)

def optimize_A_gradcpp(data, param):
    esta = EstimateAGrad()
    esta.run(data, param)
    return (data, param)

def optimize_A_grad(data, param):
    k = param["nbclust"]
    w_h = param["w_h"]
    G = data["G"]
    H = data["H"]
    S = data["S"]

    if "eta" in param:
        eta = param["eta"]
    else:
        eta = 0.1

    for j, column in enumerate(S.T):
        n1 = np.zeros(k)
        n0 = np.zeros(k)
        n = [n0,n1]
        for i, elem in enumerate(column):
            if G[i//2,j] == 1:
                n[H[i,j]][elem] += 1
            elif G[i//2,j] == 0 or G[i//2,j] == 2:
                n[H[i,j]][elem] += w_h

        for l in range(k):
            if (n1[l] + n0[l]) == 0:
                pass
            else:
                card_ik = n1[l] + n0[l]
                #data["A"][l,j] = (1 - 2*eta*card_ik)*data["A"][l,j] + 2*eta*n1[l]
                data["A"][l,j] = (1 - eta)*data["A"][l,j] + eta*n1[l]/float(card_ik)

    return (data, param)

def balance_SA(data, param):
    (n, m) = data["G"].shape
    k = param["nbclust"]
    S = data["S"]
    A = data["A"]

    def apply_assignment_S(Sj, permutation):
        inv_permutation = np.argsort(permutation)
        for i, val_s in enumerate(Sj):
            Sj[i] = inv_permutation[val_s]
        return Sj

    def apply_assignment_A(Aj, permutation):
        Aj = Aj[permutation]
        return Aj

    for j in range(1,m):
        weights = np.zeros((k,k))
        for (x,y), c in Counter(zip(S[:,j-1], S[:,j])).iteritems():
            weights[x, y] = c
        assignment = linear_assignment(-weights)
        permutation = assignment[:,1].T

        Sj = apply_assignment_S(S[:, j], permutation)
        S[:, j] = Sj
        Aj = apply_assignment_A(A[:, j], permutation)
        A[:, j] = Aj

    return (data, param)


#initialize
_init_estimatea()



import loter.estimatea as esta
import loter.estimateh as esth
import loter.graph as ests
import loter.toolsfunc as toolsfunc
import loter.metrics as metrics

import logging
import sys

root = logging.getLogger()
root.setLevel(logging.WARNING)

ch = logging.StreamHandler(sys.stderr)
ch.setLevel(logging.WARNING)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
ch.setFormatter(formatter)
root.addHandler(ch)
logger = root

"""
An "opti" is the main function solving the problem.
Taking data and param as input
Return data and param
"""
def classic_opti(data, param):
    nb_iter = param["nb_iter"]

    for i in range(nb_iter):
        data, param = ests.optimize_S(data, param)
        data, param = esta.optimize_A(data, param)
        data, param = esth.optimize_H(data, param)


    return data, param
    #return toolsfunc.repeated(toolsfunc.compose(esth.optimize_H, esta.optimize_A, ests.optimize_S),
    #                           nb_iter)((data, param))

def classic_opti_sha(data, param):
    nb_iter = param["nb_iter"]

    for i in range(nb_iter):
        data, param = ests.optimize_S(data, param)
        data, param = esth.optimize_H(data, param)
        data, param = esta.optimize_A(data, param)


    return data, param

def grad_opti(data, param):
    nb_iter = param["nb_iter"]

    for i in range(nb_iter):
        data, param = ests.optimize_S(data, param)
        data, param = esta.optimize_A_gradcpp(data, param)
        data, param = esth.optimize_H(data, param)


    return data, param


def join_opti(data, param):
    nb_iter = param["nb_iter"]

    for i in range(nb_iter):
        data, param = ests.optimize_SH(data, param)
        data, param = esth.optimize_H_old(data, param)
        data, param = esta.optimize_A(data, param)
        logger.debug("iter {}, cost : {}".format(i, metrics.cost_error_l2(data["G"],
                                                                          data["H"],
                                                                          data["S"],
                                                                          data["A"],
                                                                          param["penalty"])))

    return data, param

def join_balance(data, param):
    nb_iter = param["nb_iter"]

    for i in range(nb_iter):
        data, param = ests.optimize_SH(data, param)
        data, param = esth.optimize_H_old(data, param)
        data, param = esta.balance_SA(data, param)
        data, param = esta.optimize_A_grad(data, param)
        logger.debug("iter {}, cost before balancing: {}".format(i, metrics.cost_error_l2(data["G"],
                                                                                          data["H"],
                                                                                          data["S"],
                                                                                          data["A"],
                                                                                          param["penalty"])))
    return data, param

def join_then_classic(data, param):
    nb_iter = param["nb_iter"]

    for i in range(1):
        data, param = ests.optimize_SH(data, param)
        data, param = esth.optimize_H_old(data, param)
        data, param = esta.optimize_A(data, param)

    for i in range(nb_iter - 1):
        data, param = ests.optimize_S(data, param)
        data, param = esta.optimize_A(data, param)
        data, param = esth.optimize_H_old(data, param)

    return data, param

def join_then_classic_withsampling(data, param):
    nb_iter = param["nb_iter"]

    for i in range(1):
        data, param = ests.optimize_SH(data, param)
        data, param = esth.optimize_H(data, param)
        data, param = esta.optimize_A(data, param)

    for i in range(nb_iter - 1):
        data, param = ests.optimize_S(data, param)
        data, param = esta.optimize_A(data, param)
        data, param = esth.optimize_H(data, param)

    return data, param

def opti_with_old_H(data, param):
    nb_iter = param["nb_iter"]

    for i in range(nb_iter):
        data, param = ests.optimize_S(data, param)
        data, param = esta.optimize_A(data, param)
        data, param = esth.optimize_H_old(data, param)
        logger.debug("iter {}, cost : {}".format(i, metrics.cost_error_l2(data["G"],
                                                                          data["H"],
                                                                          data["S"],
                                                                          data["A"],
                                                                          param["penalty"])))

    return data, param

def opti_build(h, a, s):
    def optimizer(data, param):
        nb_iter = param["nb_iter"]
        return toolsfunc.repeated(toolsfunc.compose(esth.optih["opti py old fast"],
                                                    esta.optimize_A,
                                                    ests.optimize_S),
                                  nb_iter)((data, param))
    return optimizer

opti = {
    "classic_opti": classic_opti,
    "classic_opti_sha": classic_opti_sha,
    "grad_opti": grad_opti,
    "join_opti": join_opti,
    "join_balance": join_balance,
    "opti_h_old": opti_with_old_H,
    "join_then_classic": join_then_classic,
    "join_then_classic_withsampling": join_then_classic_withsampling
}



#!/usr/bin/env python

"""Comand line tool for Local Ancestry Inference (LAI) with Loter
"""

from __future__ import print_function

import os
import sys
import numpy as np
import argparse
import textwrap

import loter.locanc.local_ancestry as lc


def vcf2npy(vcfpath):
    """Convert a haplotype matrix stored in a VCF file to a Numpy array

    Args:
        vcfpath (string): complete VCF file name.

    Return: the haplotype matrix as a Numpy array.
    """
    import allel
    callset = allel.read_vcf(vcfpath)
    haplotypes_1 = callset['calldata/GT'][:,:,0]
    haplotypes_2 = callset['calldata/GT'][:,:,1]

    m, n = haplotypes_1.shape
    mat_haplo = np.empty((2*n, m))
    mat_haplo[::2] = haplotypes_1.T
    mat_haplo[1::2] = haplotypes_2.T

    return mat_haplo.astype(np.uint8)


def load_data(file_name, format="npy", verbose=True):
    """load haplotype array (from saved Numpy array or VCF files)

    Args:
        file_name (string): file storing the haplotype data.
        format (string): data format 'npy' for saved Numpy array, 'vcf' for
            VCF data file. Default is 'npy'.
        verbose (bool): set verbosity. Default is True.

    Return: the haplotype matrix as a Numpy array.
    """
    if verbose:
        print("loading file {}".format(file_name))
    data_array = None
    if format == "npy":
        data_array = np.load(os.path.expanduser(file_name), allow_pickle = True)
    elif format == "txt":
        data_array = np.genfromtxt(os.path.expanduser(file_name), dtype=np.uint8)
    elif format == "vcf":
        try:
            data_array = vcf2npy(file_name)
        except (ModuleNotFoundError):
            print("Using VCF files requires the `scikit-allel` package "
                  + "(available with `pip install scikit-allel`)")
    else:
        raise ValueError("Wrong data file format")
    return data_array.astype(np.uint8)


def main():
    """Command line interface for Loter
    """
    ## args
    parser = argparse.ArgumentParser(
        prog="loter_cli",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description = __doc__,
        epilog=textwrap.dedent('''\
            **Note:**
                Haplotype input matrices should be organised
                as follows: with haplotypes (samples) in rows and SNPs
                in columns. Ancestries of admixed haplotypes inferred by Loter
                will be stored in the same way.
            
            **Important:** When using text format (csv) for input data, 
            missing values should be encoded as 255 or NA.

            Example (run from Loter project root directory):
            loter_cli -r data/H_ceu.npy data/H_yri.npy -a data/H_mex.npy -f npy -o tmp.npy -n 8 -v'''))
    requiredNamed = parser.add_argument_group('required named arguments')
    requiredNamed.add_argument(
                        "-r", "--ref", nargs='+',
                        help="files storing input reference haplotypes.",
                        required=True)
    requiredNamed.add_argument(
                        "-a", "--adm", nargs=None,
                        help="file storing input admixed haplotypes.",
                        required=True)
    parser.add_argument("-f", "--format", default="npy", nargs=None,
                        help="file format: `npy` (default) for saved numpy arrays, "
                             + "`txt` for csv text files (experimental) "
                             + "or `vcf` for VCF files (which require "
                             + "the `scikit-allel` package). "
                             + "**Note:** all input files (references and admixed) "
                             + "should have the same format.")
    requiredNamed.add_argument(
                        "-o", "--output", nargs=None,
                        help="file to store the infered ancestries of admixed haplotypes, "
                             + "saved as a numpy array by default, or as a text CSV file "
                             + " if the file extension is '.txt'.", required=True)
    parser.add_argument("-n", "--ncore", default=1, type=int, nargs=None,
                        help="Number of CPU cores to use. Default is 1.")
    parser.add_argument("-pc", "--phase_correction",
                        help="Run the phase correction module after the "
                             + "inference algorithm (only available for "
                             + "data from diploid organisms).",
                        action="store_true")
    parser.add_argument("-v", "--verbose", help="Set verbosity on.",
                        action="store_true")
    args = parser.parse_args()

    # set verbosity
    VERBOSE = args.verbose
    def pprint(string):
        """custom print function
        """
        if VERBOSE:
            print(string)

    ## data
    pprint("## Loading reference haplotypes")
    pprint("Using {} reference populations".format(len(args.ref)))
    H_ref = []
    for index, file in enumerate(args.ref):
        H_ref.append(load_data(file, args.format, args.verbose))
        pprint("Dimension of reference haplotype matrix {} = {}"
                    .format(index, H_ref[-1].shape))

    pprint("## Loading admixed haplotypes")
    H_adm = load_data(args.adm, args.format)
    pprint("Dimension of admixed haplotype matrix = {}"
                .format(H_adm.shape))

    pprint("## Run loter")
    res_loter = None
    if args.phase_correction:
        pprint("Using Loter with bagging and phase correction")
        res_loter = lc.loter_smooth(
                        l_H=H_ref, h_adm=H_adm, range_lambda=np.arange(1.5, 5.5, 0.5),
                        threshold=0.90, rate_vote=0.5, nb_bagging=20,
                        num_threads=args.ncore)
    else:
        pprint("Using Loter with bagging")
        res_loter = lc.loter_local_ancestry(
                        l_H=H_ref, h_adm=H_adm, range_lambda=np.arange(1.5, 5.5, 0.5),
                        rate_vote=0.5, nb_bagging=20, num_threads=args.ncore)
        res_loter = res_loter[0]

    pprint("## saving result to file {}".format(args.output))
    if args.output.lower().endswith('.txt'):
        np.savetxt(os.path.expanduser(args.output), res_loter, fmt="%i")
    else:
        np.save(os.path.expanduser(args.output), res_loter)


if __name__ == "__main__":
    main()



import numpy as np

# mode function from scipy to remove one dependency
def mode(a, axis=0):
    scores = np.unique(np.ravel(a))       # get ALL unique values
    testshape = list(a.shape)
    testshape[axis] = 1
    oldmostfreq = np.zeros(testshape)
    oldcounts = np.zeros(testshape)

    for score in scores:
        template = (a == score)
        counts = np.expand_dims(np.sum(template, axis),axis)
        mostfrequent = np.where(counts > oldcounts, score, oldmostfreq)
        oldcounts = np.maximum(counts, oldcounts)
        oldmostfreq = mostfrequent

    return mostfrequent, oldcounts

def build_SA(S, A):
    _, m = S.shape
    x = np.arange(m)
    # x will be broadcast to the size of S
    return A[S, x]




from setuptools import find_packages
from distutils.core import setup, Extension
import distutils.util
import os
import sys
from glob import glob

## path
curr_path = os.path.dirname(os.path.abspath(os.path.expanduser(__file__)))
src_path = os.path.join(curr_path, 'src')

## src files
src = []
pattern = '*.cpp'
for subdir,_,_ in os.walk(src_path):
    src.extend(glob(os.path.join(subdir, pattern)))

## dependencies
dep = []
patterns = ['*.h', '*.hpp']
for subdir,_,_ in os.walk(src_path):
    for pattern in patterns:
        dep.extend(glob(os.path.join(subdir, pattern)))


## include and linking
incs = []
libdirs = []
libs = []
link_flags = []

## compile flags
cxx_flags = ['-Wall', '-fPIC', '-O3', '-msse2']

if "--no_openmp" in sys.argv:
    cxx_flags.append('-DDISABLE_OPENMP')
    sys.argv.remove("--no_openmp")
    libs = ['stdc++', 'blas', 'lapack']
else:
    cxx_flags.append('-fopenmp')
    link_flags.append('-fopenmp')
    libs += ['stdc++', 'blas', 'lapack', 'gomp']

## C++ extension
loter_wrap = Extension(
    'loter._loter_wrap',
    sources = src,
    include_dirs = incs,
    extra_compile_args = cxx_flags,
    library_dirs = libdirs,
    libraries = libs,
    extra_link_args = link_flags,
    language = 'c++',
    depends = dep,
)


setup(
    name='loter',
    version='1.0',
    description="Optimization Package to phase haplotypes",
    author="Thomas Dias-Alves",
    url = 'https://github.com/bcm-uga/Loter',
    license = "MIT",
    license_files = ('LICENSE.txt',),
    setup_requires=['numpy'],
    install_requires=[
        "numpy",
        "pandas",
        "scikit-learn",
        "scipy"
    ],
    ext_modules=[loter_wrap,],
    packages=find_packages(),
    entry_points={
        'console_scripts': ['loter_cli=loter.cli:main'],
    },
    data_files=[],
    zip_safe=False
)



import ctypes as C
import numpy as np

import loter.errorhandler as errorhandler
import loter.datastruct.parameter as parameter
from loter.find_lib import _LIB

@errorhandler.eh_fn
def EstimateSErrorHandlerFn(error_message, user_data):
    """Callback function for C api in errorhandler.c

    """
    print("Error Estimation of S")
    print(error_message)

class EstimateSException(Exception):
    """Exception raised for errors in EstimateS calls

    Attributes:
        value -- input in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, value, message):
        self.value = value
        self.message = message

def _init_estimates():

    # Constructor
    _LIB.estimates_create.argtypes = [
        C.POINTER(errorhandler.ErrorHandler)
    ]
    _LIB.estimates_create.restype = C.c_void_p

    # Destructor
    _LIB.estimates_destroy.argtypes = [
        C.c_void_p,
        C.POINTER(errorhandler.ErrorHandler)
    ]
    _LIB.estimates_destroy.restype = None

    # Getter Penalty
    _LIB.estimates_run.argtypes = [
        C.c_void_p,
        np.ctypeslib.ndpointer(dtype = np.uint8,
                               ndim=2,
                               flags='C_CONTIGUOUS'),
        np.ctypeslib.ndpointer(dtype = np.uint8,
                               ndim=2,
                               flags='C_CONTIGUOUS'),
        np.ctypeslib.ndpointer(dtype = np.float32,
                               ndim=2,
                               flags='C_CONTIGUOUS'),
        np.ctypeslib.ndpointer(dtype = np.uint8,
                               ndim=2,
                               flags='C_CONTIGUOUS'),
        C.c_int,
        C.c_int,
        C.c_void_p,
        C.POINTER(errorhandler.ErrorHandler)
    ]
    _LIB.estimates_run.restype = None

def _init_estimatesh():

    # Constructor
    _LIB.estimatesh_create.argtypes = [
        C.POINTER(errorhandler.ErrorHandler)
    ]
    _LIB.estimatesh_create.restype = C.c_void_p

    # Destructor
    _LIB.estimatesh_destroy.argtypes = [
        C.c_void_p,
        C.POINTER(errorhandler.ErrorHandler)
    ]
    _LIB.estimatesh_destroy.restype = None

    # Getter Penalty
    _LIB.estimatesh_run.argtypes = [
        C.c_void_p,
        np.ctypeslib.ndpointer(dtype = np.uint8,
                               ndim=2,
                               flags='C_CONTIGUOUS'),
        np.ctypeslib.ndpointer(dtype = np.uint8,
                               ndim=2,
                               flags='C_CONTIGUOUS'),
        np.ctypeslib.ndpointer(dtype = np.float32,
                               ndim=2,
                               flags='C_CONTIGUOUS'),
        np.ctypeslib.ndpointer(dtype = np.uint8,
                               ndim=2,
                               flags='C_CONTIGUOUS'),
        C.c_int,
        C.c_int,
        C.c_void_p,
        C.POINTER(errorhandler.ErrorHandler)
    ]
    _LIB.estimatesh_run.restype = None

def _init_estimateshknn():

    # Constructor
    _LIB.estimateshknn_create.argtypes = [
        C.POINTER(errorhandler.ErrorHandler)
    ]
    _LIB.estimateshknn_create.restype = C.c_void_p

    # Destructor
    _LIB.estimateshknn_destroy.argtypes = [
        C.c_void_p,
        C.POINTER(errorhandler.ErrorHandler)
    ]
    _LIB.estimateshknn_destroy.restype = None

    # Getter Penalty
    _LIB.estimateshknn_run.argtypes = [
        C.c_void_p,
        np.ctypeslib.ndpointer(dtype = np.uint8,
                               ndim=2,
                               flags='C_CONTIGUOUS'),
        np.ctypeslib.ndpointer(dtype = np.uint8,
                               ndim=2,
                               flags='C_CONTIGUOUS'),
        np.ctypeslib.ndpointer(dtype = np.float32,
                               ndim=2,
                               flags='C_CONTIGUOUS'),
        np.ctypeslib.ndpointer(dtype = np.uint8,
                               ndim=2,
                               flags='C_CONTIGUOUS'),
        C.c_int,
        C.c_int,
        C.c_float,
        C.c_int,
        C.c_void_p,
        C.POINTER(errorhandler.ErrorHandler)
    ]
    _LIB.estimatesh_run.restype = None

def _init_estimatesknn():

    # Constructor
    _LIB.estimatesknn_create.argtypes = [
        C.POINTER(errorhandler.ErrorHandler)
    ]
    _LIB.estimatesknn_create.restype = C.c_void_p

    # Destructor
    _LIB.estimatesknn_destroy.argtypes = [
        C.c_void_p,
        C.POINTER(errorhandler.ErrorHandler)
    ]
    _LIB.estimatesknn_destroy.restype = None

    # Getter Penalty
    _LIB.estimatesknn_run.argtypes = [
        C.c_void_p,
        np.ctypeslib.ndpointer(dtype = np.uint8,
                               ndim=2,
                               flags='C_CONTIGUOUS'),
        np.ctypeslib.ndpointer(dtype = np.uint8,
                               ndim=2,
                               flags='C_CONTIGUOUS'),
        np.ctypeslib.ndpointer(dtype = np.uint32,
                               ndim=2,
                               flags='C_CONTIGUOUS'),
        np.ctypeslib.ndpointer(dtype = np.float32,
                               ndim=1,
                               flags='C_CONTIGUOUS'),
        C.c_int,
        C.c_int,
        C.c_int,
        C.c_void_p,
        C.POINTER(errorhandler.ErrorHandler)
    ]
    _LIB.estimatesknn_run.restype = None

class EstimateS(object):

    def __init__(self):
        self._EH = errorhandler.ErrorHandler(EstimateSErrorHandlerFn, None)
        self.obj = _LIB.estimates_create(C.byref(self._EH))

    def run(self, data, param):

        param_opti = parameter.ParameterOptimization(param["penalty"],
                                                     param["nbclust"],
                                                     param["w_h"],
                                                     param["nb_iter"])

        G, H, A, S = data["G"], data["H"], data["A"], data["S"]
        n, m = G.shape

        _LIB.estimates_run(self.obj,
                           G,
                           H,
                           A,
                           S,
                           n,
                           m,
                           param_opti.obj,
                           C.byref(self._EH))

    def __del__(self):
        _LIB.estimates_destroy(self.obj, C.byref(self._EH))

def optimize_S(data, param):
    ests = EstimateS()
    ests.run(data, param)
    return (data, param)

class EstimateSH(object):

    def __init__(self):
        self._EH = errorhandler.ErrorHandler(EstimateSErrorHandlerFn, None)
        self.obj = _LIB.estimatesh_create(C.byref(self._EH))

    def run(self, data, param):
        param_opti = parameter.ParameterOptimization(param["penalty"],
                                                     param["nbclust"],
                                                     param["w_h"],
                                                     param["nb_iter"])

        G, H, A, S = data["G"], data["H"], data["A"], data["S"]
        n, m = G.shape

        _LIB.estimatesh_run(self.obj,
                            G,
                            H,
                            A,
                            S,
                            n,
                            m,
                            param_opti.obj,
                            C.byref(self._EH))

    def __del__(self):
        _LIB.estimatesh_destroy(self.obj, C.byref(self._EH))

def optimize_SH(data, param):
    ests = EstimateSH()
    ests.run(data, param)
    return (data, param)

class EstimateSHknn(object):

    def __init__(self):
        self._EH = errorhandler.ErrorHandler(EstimateSErrorHandlerFn, None)
        self.obj = _LIB.estimateshknn_create(C.byref(self._EH))

    def run(self, data, param):
        param_opti = parameter.ParameterOptimization(param["penalty"],
                                                     param["nbclust"],
                                                     param["w_h"],
                                                     param["nb_iter"])

        G, H, A, S = data["G"], data["H"], data["A"], data["S"]
        n, m = G.shape

        _LIB.estimateshknn_run(self.obj,
                               G,
                               H,
                               A,
                               S,
                               n,
                               m,
                               param["small_penalty"],
                               param["num_threads"],
                               param_opti.obj,
                               C.byref(self._EH))

    def __del__(self):
        _LIB.estimateshknn_destroy(self.obj, C.byref(self._EH))

def optimize_SHknn(data, param):
    ests = EstimateSHknn()
    ests.run(data, param)
    return (data, param)

class EstimateSknn(object):

    def __init__(self):
        self._EH = errorhandler.ErrorHandler(EstimateSErrorHandlerFn, None)
        self.obj = _LIB.estimatesknn_create(C.byref(self._EH))

    def run(self, data, param):

        param_opti = parameter.ParameterOptimization(param["penalty"],
                                                     param["nbclust"],
                                                     param["w_h"],
                                                     param["nb_iter"])

        G, H, A, S = data["G"], data["H"], data["A"], data["S"]
        n, m = G.shape

        _LIB.estimatesknn_run(self.obj,
                              np.ascontiguousarray(H.T),
                              np.ascontiguousarray(A.T),
                              np.ascontiguousarray(S),
                              np.ascontiguousarray(param["weights"]).astype(np.float32),
                              n,
                              m,
                              param["num_threads"],
                              param_opti.obj,
                              C.byref(self._EH))

    def __del__(self):
        _LIB.estimatesknn_destroy(self.obj, C.byref(self._EH))

def optimize_Sknn(data, param):
    estsknn = EstimateSknn()
    estsknn.run(data, param)
    return (data, param)

#initialize
_init_estimates()
_init_estimatesh()
_init_estimatesknn()
_init_estimateshknn()



import ctypes as C

import loter.errorhandler as errorhandler
from loter.find_lib import _LIB

@errorhandler.eh_fn
def ParameterOptimizationErrorHandlerFn(error_message, user_data):
    """Callback function for C api in errorhandler.c

    """
    print("Error ParameterOptimization")
    print(error_message)

class ParameterOptimizationException(Exception):
    """Exception raised for errors in input parameters

    Attributes:
        value -- input in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, value, message):
        self.value = value
        self.message = message

def _init_parameter_optimization():

    # Constructor
    _LIB.parameterOptimization_create.argtypes = [
        C.c_float,
        C.c_int,
        C.c_float,
        C.c_int,
        C.POINTER(errorhandler.ErrorHandler)
    ]
    _LIB.parameterOptimization_create.restype = C.c_void_p

    # Destructor
    _LIB.parameterOptimization_destroy.argtypes = [
        C.c_void_p,
        C.POINTER(errorhandler.ErrorHandler)
    ]
    _LIB.parameterOptimization_destroy.restype = None

    # Getter Penalty
    _LIB.parameterOptimization_get.argtypes = [
        C.c_void_p,
        C.c_char_p,
        C.POINTER(errorhandler.ErrorHandler)
    ]
    _LIB.parameterOptimization_get.restype = C.c_char_p
    
    # Setter Penalty
    _LIB.parameterOptimization_set.argtypes = [
        C.c_void_p,
        C.c_char_p,
        C.c_char_p,
        C.POINTER(errorhandler.ErrorHandler)
    ]
    _LIB.parameterOptimization_set.restype = None

class ParameterOptimization(object):

    def __init__(self, penalty, nbclust, w_h, nb_iter):
        self._EH = errorhandler.ErrorHandler(ParameterOptimizationErrorHandlerFn, None)

        self.obj = _LIB.parameterOptimization_create(float(penalty),
                                                     int(nbclust),
                                                     float(w_h),
                                                     int(nb_iter),
                                                     C.byref(self._EH))
    @property
    def penalty(self):
        return float(_LIB.parameterOptimization_get(self.obj,
                                                    "penalty",
                                                    C.byref(self._EH)))

    @penalty.setter
    def penalty(self, value):
        if value < 0:
            raise ParameterOptimizationException(value, "Penalty < 0")
        _LIB.parameterOptimization_set(self.obj,
                                       "penalty",
                                       str(value),
                                       C.byref(self._EH))
 
    @property
    def nbclust(self):
        return int(_LIB.parameterOptimization_get(self.obj,
                                                  "nbclust",
                                                  C.byref(self._EH)))

    @nbclust.setter
    def nbclust(self, value):
        if value < 0:
            raise ParameterOptimizationException(value, "Number of clusters < 0")
        _LIB.parameterOptimization_set(self.obj,
                                       "nbclust",
                                       str(value),
                                       C.byref(self._EH))

    @property
    def w_h(self):
        return float(_LIB.parameterOptimization_get(self.obj,
                                                    "w_h",
                                                    C.byref(self._EH)))

    @w_h.setter
    def w_h(self, value):
        if value < 0:
            raise ParameterOptimizationException(value, "Weight of homozygous < 0")
        _LIB.parameterOptimization_set(self.obj,
                                       "w_h",
                                       str(value),
                                       C.byref(self._EH))
               
    @property
    def nb_iter(self):
        return int(_LIB.parameterOptimization_get(self.obj,
                                                  "nb_iter",
                                                  C.byref(self._EH)))

    @nb_iter.setter
    def nb_iter(self, value):
        if value <= 0:
            raise ParameterOptimizationException(value, "Number of Iterations <= 0")
        _LIB.parameterOptimization_set(self.obj,
                                       "nb_iter",
                                       str(value),
                                       C.byref(self._EH))
 
    def __del__(self):
        _LIB.parameterOptimization_destroy(self.obj, self._EH)

#initialize
_init_parameter_optimization()



def compose(*functions):
    def inner(arg):
        for f in reversed(functions):
            arg = f(*arg)
        return arg
    return inner

def repeated(function, n):
    def r_function(arg):
        for _ in xrange(n):
            arg = function(arg)
        return arg
    return r_function



import ctypes as C
import numpy as np

import pyhaplophase.methods.SAH_opti as sah
import pyhaplophase.dataset as dataset
import pyhaplophase.parametersopti as paramsopti
import pyhaplophase.utils as utils
import pyhaplophase.error as error
import pyhaplophase.experiments.experiment as experiment
import pyhaplophase.dumper.cout_dumper as coutdumper
import cProfile
import os
import loter.metrics as metrics
import loter.predcombine as combine
import loter.pipeline as pipeline

from loter.find_lib import _LIB

def wrapper_cpp(G, H, A, S, nb_iter, w, penalty):
    (n, m) = G.shape
    (k, _) = A.shape

    f_eigen_loter = _LIB.haplophase_cpp
    f_eigen_loter.argtypes = [np.ctypeslib.ndpointer(dtype = np.uint8,
                                                          ndim=2,
                                                          flags='C_CONTIGUOUS'),
                                   np.ctypeslib.ndpointer(dtype = np.uint8,
                                                          ndim=2,
                                                          flags='C_CONTIGUOUS'),
                                   np.ctypeslib.ndpointer(dtype = np.float32,
                                                          ndim=2,
                                                          flags='C_CONTIGUOUS'),
                                   np.ctypeslib.ndpointer(dtype = np.uint8,
                                                          ndim=2,
                                                          flags='C_CONTIGUOUS'),
                                   C.c_int, C.c_int, C.c_int, C.c_int,
                                   C.c_float, C.c_float]
    f_eigen_loter.restype = None

    return f_eigen_loter(G, H, A, S, n, m, k, nb_iter, w, penalty)

def wrapper_all(G, k, nb_iter, nb_run, w, penalty):
    (n, m) = G.shape

    f_eigen_loter = _LIB.haplophase_all
    f_eigen_loter.argtypes = [np.ctypeslib.ndpointer(dtype = np.uint8,
                                                          ndim=2,
                                                          flags='C_CONTIGUOUS'),
                                   np.ctypeslib.ndpointer(dtype = np.uint8,
                                                          ndim=2,
                                                          flags='C_CONTIGUOUS'),
                                   C.c_int, C.c_int, C.c_int, C.c_int, C.c_int,
                                   C.c_float, C.c_float]
    f_eigen_loter.restype = None

    H = np.zeros((2*n,m), dtype=np.uint8)
    f_eigen_loter(G, H, n, m, k, nb_iter, nb_run, w, penalty)
    return H

def wrapper_parallel(G, k, nb_iter, nb_run, w, penalty, num_threads):
    (n, m) = G.shape

    f_eigen_loter = _LIB.haplophase_parallel
    f_eigen_loter.argtypes = [np.ctypeslib.ndpointer(dtype = np.uint8,
                                                          ndim=2,
                                                          flags='C_CONTIGUOUS'),
                                   np.ctypeslib.ndpointer(dtype = np.uint8,
                                                          ndim=2,
                                                          flags='C_CONTIGUOUS'),
                                   C.c_int, C.c_int, C.c_int, C.c_int, C.c_int,
                                   C.c_float, C.c_float, C.c_int]
    f_eigen_loter.restype = None

    H = np.zeros((2*n,m), dtype=np.uint8)
    f_eigen_loter(G, H, n, m, k, nb_iter, nb_run, w, penalty, num_threads)
    return H

def wrapper_parallel_lambdvar(G, lambdvar, k, nb_iter, nb_run, w, penalty, num_threads):
    (n, m) = G.shape
    nb_lambdvar = len(lambdvar)

    f_eigen_loter = _LIB.haplophase_lambdvar
    f_eigen_loter.argtypes = [np.ctypeslib.ndpointer(dtype = np.uint8,
                                                     ndim=2,
                                                     flags='C_CONTIGUOUS'),
                              np.ctypeslib.ndpointer(dtype = np.uint8,
                                                     ndim=2,
                                                     flags='C_CONTIGUOUS'),
                              np.ctypeslib.ndpointer(dtype = np.uint8,
                                                     ndim=1,
                                                     flags='C_CONTIGUOUS'),
                              C.c_int,
                              C.c_int, C.c_int, C.c_int, C.c_int, C.c_int,
                              C.c_float, C.c_float, C.c_int]
    f_eigen_loter.restype = None

    H = np.zeros((2*n,m), dtype=np.uint8)
    f_eigen_loter(G, H, lambdvar, nb_lambdvar, n, m, k, nb_iter, nb_run, w, penalty, num_threads)
    return H

def create_data():
    G = np.array([[2,2,0,2],
                  [1,1,0,1]])
    H = np.array([[1,1,0,1],[1,1,0,1],[0,1,0,1],[1,0,0,0]])
    A = np.array([[0.1,0.5,0.8,0.9],
                  [0.0,0.1,0.2,0.9]], dtype=np.float32)
    S = np.array([[0,0,0,0],[0,0,0,0],[1,1,1,1],[0,0,0,0]])

    G = np.ascontiguousarray(G).astype(np.uint8)
    H = np.ascontiguousarray(H).astype(np.uint8)
    S = np.ascontiguousarray(S).astype(np.uint8)

    return {"G":G,"H":H,"A":A,"S":S}

def create_param():
    return {"penalty": 2.0,
            "nbclust": 2,
            "w_h": 100.0,
            "nb_iter": 10}

if __name__ == '__main__':
    print("Main wrapper")
